""""
Main module that serves as an entry point for Vakt decisions.
Also contains Inquiry class.
"""

import logging

from .util import JsonSerializer, PrettyPrint


log = logging.getLogger(__name__)


class Inquiry(JsonSerializer, PrettyPrint):
    """Holds all the information about the inquired intent.
    Is responsible to decisions if the inquired intent allowed or not."""

    def __init__(self, resource=None, action=None, subject=None, context=None):
        # explicitly assign empty strings instead of occasional None, (), etc.
        self.resource = resource or ''
        self.action = action or ''
        self.subject = subject or ''
        self.context = context or {}

    @classmethod
    def from_json(cls, data):
        props = cls._parse(data)
        return cls(**props)

    def to_json_sorted(self):
        """
        Get JSON representation with all keys sorted.
        """
        return super().to_json(sort=True)

    def __eq__(self, other):
        """
        If inquiries have the same contents - they are equal
        """
        return self.to_json_sorted() == other.to_json_sorted()

    def __hash__(self):
        """
        We do not use to_json as contents representation, because strings are not guaranteed
        to be hashed consistently across different python processes.
        """
        return hash(tuple((ord(c) for c in self.to_json_sorted())))


class Guard:
    """
    Executor of policy checks.
    Given a storage and a checker it can decide via `is_allowed` method if a given inquiry allowed or not.
    """

    def __init__(self, storage, checker):
        self.storage = storage
        self.checker = checker

    def is_allowed(self, inquiry):
        """
        Is given inquiry intent allowed or not?
        Same as `is_allowed_no_audit`, but also logs policy enforcement decisions to audit-log.
        Is meant to be used by an end-user.
        """
        answer = self.is_allowed_no_audit(inquiry)
        if answer:
            log.info('Incoming Inquiry was allowed. Inquiry: %s', inquiry)
        else:
            log.info('Incoming Inquiry was rejected. Inquiry: %s', inquiry)
        return answer

    def is_allowed_no_audit(self, inquiry):
        """
        Is given inquiry intent allowed or not?
        Does not log answers.
        Is not meant to be called by an end-user. Use it only if you want the core functionality of allowance check.
        """
        try:
            policies = self.storage.find_for_inquiry(inquiry, self.checker)
            # Storage is not obliged to do the exact policies match. It's up to the storage
            # to decide what policies to return. So we need a more correct programmatically done check.
            answer = self.check_policies_allow(inquiry, policies)
        except Exception:
            log.exception('Unexpected exception occurred while checking Inquiry %s', inquiry)
            answer = False
        return answer

    def check_policies_allow(self, inquiry, policies):
        """
        Check if any of a given policy allows a specified inquiry
        """
        # If no policies found or None is given -> deny access!
        if not policies:
            return False

        # Filter policies that fit Inquiry by its attributes.
        filtered = [p for p in policies if
                    self.checker.fits(p, 'actions', inquiry.action, inquiry) and
                    self.checker.fits(p, 'subjects', inquiry.subject, inquiry) and
                    self.checker.fits(p, 'resources', inquiry.resource, inquiry) and
                    self.check_context_restriction(p, inquiry)]

        # no policies -> deny access!
        # if we have 2 or more similar policies - all of them should have allow effect, otherwise -> deny access!
        return len(filtered) > 0 and all(p.allow_access() for p in filtered)

    @staticmethod
    def check_context_restriction(policy, inquiry):
        """
        Check if context restriction in the policy is satisfied for a given inquiry's context.
        If at least one rule is not present in Inquiry's context -> deny access.
        If at least one rule provided in Inquiry's context is not satisfied -> deny access.
        """
        for key, rule in policy.context.items():
            try:
                ctx_value = inquiry.context[key]
            except KeyError:
                log.debug("No key '%s' found in Inquiry context", key)
                return False
            if not rule.satisfied(ctx_value, inquiry):
                return False
        return True



import json
import re
import unittest

import pytest

from vakt.storage.mongo import *
from vakt.rules.base import Rule
from vakt.guard import Inquiry, Guard
from vakt import version_info
from .test_mongo import DB_NAME, COLLECTION, MIGRATION_COLLECTION, create_client


@pytest.mark.integration
class TestMongoMigrationSet:

    @pytest.fixture()
    def migration_set(self):
        client = create_client()
        storage = MongoStorage(client, DB_NAME, collection=COLLECTION)
        yield MongoMigrationSet(storage, MIGRATION_COLLECTION)
        client[DB_NAME][COLLECTION].delete_many({})
        client[DB_NAME][MIGRATION_COLLECTION].delete_many({})
        client.close()

    def test_application_of_migration_number(self, migration_set):
        assert 0 == migration_set.last_applied()
        migration_set.save_applied_number(6)
        assert 6 == migration_set.last_applied()
        migration_set.save_applied_number(2)
        assert 2 == migration_set.last_applied()

    def test_up_and_down(self, migration_set):
        migration_set.save_applied_number(0)
        migration_set.up()
        assert 4 == migration_set.last_applied()
        migration_set.up()
        assert 4 == migration_set.last_applied()
        migration_set.down()
        assert 0 == migration_set.last_applied()
        migration_set.down()
        assert 0 == migration_set.last_applied()


@pytest.mark.integration
class TestMigration0To1x0x3:

    @pytest.fixture()
    def migration(self):
        client = create_client()
        storage = MongoStorage(client, DB_NAME, collection=COLLECTION)
        yield Migration0To1x1x0(storage)
        client[DB_NAME][COLLECTION].delete_many({})
        client.close()

    def test_order(self, migration):
        assert 1 == migration.order

    def test_has_access_to_storage(self, migration):
        assert hasattr(migration, 'storage') and migration.storage is not None

    def test_up(self, migration):
        migration.up()
        created_indices = [i['name'] for i in migration.storage.collection.list_indexes()]
        assert created_indices == ['_id_', 'actions_idx', 'subjects_idx', 'resources_idx']

    def test_down(self, migration):
        migration.down()
        left_indices = [i['name'] for i in migration.storage.collection.list_indexes()]
        assert left_indices == ['_id_']


# Custom classes for TestMigration1x1x0To1x1x1
class WithObject(Rule):  # pragma: no cover
    def __init__(self, val):
        self.val = re.compile(val)

    def satisfied(self, what=None, inquiry=None):
        return True


class Simple(Rule):   # pragma: no cover
    def __init__(self, val):
        self.val = val

    def satisfied(self, what=None, inquiry=None):
        return True


@pytest.mark.integration
class TestMigration1x1x0To1x1x1:
    class MockLoggingHandler(logging.Handler):
        def __init__(self, *args, **kwargs):
            self.messages = {}
            super().__init__(*args, **kwargs)

        def emit(self, record):
            level = record.levelname.lower()
            if level not in self.messages:
                self.messages[level] = []
            self.messages[level].append(record.getMessage())

    @pytest.fixture()
    def storage(self):
        client = create_client()
        storage = MongoStorage(client, DB_NAME, collection=COLLECTION)
        yield storage
        client[DB_NAME][COLLECTION].delete_many({})
        client.close()

    def test_order(self, storage):
        migration = Migration1x1x0To1x1x1(storage)
        assert 2 == migration.order

    def test_up(self, storage):
        migration = Migration1x1x0To1x1x1(storage)
        # prepare docs that might have been saved by users in v 1.1.0
        docs = [
            (
                """
                { "_id" : 10, "uid" : 10, "description" : null, "subjects" : [ ], "effect" : "allow", 
                "resources" : [ ], "actions" : [ ], "rules" : { "secret" : 
                "{\\"type\\": \\"vakt.rules.string.StringEqualRule\\", \\"contents\\": {\\"val\\": \\"i-am-a-foo\\"}}", 
                "name":"{\\"type\\": \\"vakt.rules.string.StringEqualRule\\", \\"contents\\":{\\"val\\": \\"Max\\"}}" }}
                """,
                """
                { "_id" : 10, "actions" : [ ], "description" : null, "effect" : "allow", "resources" : [ ], 
                "rules" : { "name" : {"py/object": "vakt.rules.string.StringEqualRule", "val": "Max" },
                "secret" : {"py/object": "vakt.rules.string.StringEqualRule", "val": "i-am-a-foo"} }, 
                "subjects" : [ ], "uid" : 10 }
                """
            ),
            (
                """
                { "_id" : 20, "uid" : 20, "description" : "foo bar", "subjects" : [ "<.*>" ], "effect" : "allow",
                "resources" : [ "<.*>" ], "actions" : [ "<.*>" ], "rules" : { "secret" :
                "{\\"type\\": \\"vakt.rules.string.StringEqualRule\\", \\"contents\\": {\\"val\\": \\"John\\"}}" } }
                """,
                """
                { "_id" : 20, "actions" : [ "<.*>" ], "description" : "foo bar", "effect" : "allow",
                "resources" : [ "<.*>" ], "rules" : { "secret" :
                { "py/object": "vakt.rules.string.StringEqualRule", "val": "John"} }, "subjects" : [ "<.*>" ],
                "uid" : 20 }
                """
            ),
            (
                """
                { "_id" : 30, "uid" : 30, "description" : "foo bar", "subjects" : [ "<.*>" ], "effect" : "allow",
                "resources" : [ "<.*>" ], "actions" : [ "<.*>" ], "rules" : {  } }
                """,
                """
                { "_id" : 30, "actions" : [ "<.*>" ], "description" : "foo bar", "effect" : "allow",
                "resources" : [ "<.*>" ], "rules" : {  }, "subjects" : [ "<.*>" ], "uid" : 30 }
                """
            ),
            (
                """
                { "_id" : 40, "uid" : 40, "description" : null, "subjects" : [ "<.*>" ], "effect" : "allow", 
                "resources" : [ "<.*>" ], "actions" : [ "<.*>" ], "rules" : 
                { "num" : 
                "{\\"type\\": \\"storage.test_mongo_migration.Simple\\", \\"contents\\": {\\"val\\": \\"123\\"}}", 
                "a" : "{\\"type\\": \\"vakt.rules.string.StringEqualRule\\", \\"contents\\": {\\"val\\": \\"foo\\"}}" }}
                """,
                """
                { "_id" : 40, "actions" : [ "<.*>" ], "description" : null, "effect" : "allow",  "resources" : ["<.*>"],
                "rules" : { "a" : {"py/object": "vakt.rules.string.StringEqualRule", "val": "foo"},
                "num" : { "py/object": "storage.test_mongo_migration.Simple", "val": "123"} }, "subjects" : ["<.*>"],
                "uid" : 40 }
                """
            ),
            (
                """
                { "_id" : 50, "uid" : 50, "description" : null, "subjects" : [ ], "effect" : "allow", "resources" : [ ], 
                "actions" : [ ], "rules" : { "num" : 
                "{\\"type\\": \\"storage.test_mongo_migration.Simple\\", \\"contents\\": {\\"val\\": \\"46\\"}}" } }
                """,
                """
                { "_id" : 50, "actions" : [ ], "description" : null, "effect" : "allow", "resources" : [ ],
                "rules" : { "num" : { "py/object": "storage.test_mongo_migration.Simple", "val": "46"} },
                "subjects" : [ ], "uid" : 50 }
                """
            ),
        ]
        for (doc, _) in docs:
            d = b_json.loads(doc)
            migration.storage.collection.insert_one(d)

        migration.up()

        # test no new docs were added and no docs deleted
        assert len(docs) == len(list(migration.storage.collection.find({})))
        # test Policy.from_json() is called without errors for each doc (implicitly)
        assert len(docs) == len(list(migration.storage.get_all(1000, 0)))
        # test string contents of each doc
        for (doc, result_doc) in docs:
            new_doc = migration.storage.collection.find_one({'uid': json.loads(doc)['uid']})
            expected = result_doc.replace("\n", '').replace(' ', '')
            actual = json.dumps(new_doc, sort_keys=True).replace("\n", '').replace(' ', '')
            assert expected == actual
        # test full guard allowance run
        if version_info() < (1, 2, 0):  # pragma: no cover
            g = Guard(migration.storage, RegexChecker())
            inq = Inquiry(action='foo', resource='bar', subject='Max', context={'val': 'foo', 'num': '123'})
            assert g.is_allowed(inq)

    def test_down(self, storage):
        assertions = unittest.TestCase('__init__')
        migration = Migration1x1x0To1x1x1(storage)
        # prepare docs that might have been saved by users in v 1.1.1
        docs = [
            (
                """
                {"_id":1,"actions":[],"description":null,"effect":"deny","resources":[],"rules":
                {"name":{"py/object":"vakt.rules.string.RegexMatchRule",
                "regex":{"pattern":"[Mm]ax","py/object":"_sre.SRE_Pattern"}},
                "secret":{"py/object":"vakt.rules.string.StringEqualRule","val":"i-am-a-foo"}},
                "subjects":[],"uid":1}
                """,
                None
            ),
            (
                """
                {"_id":2,"actions":["<.*>"],"description":"foobar","effect":"deny",
                "resources":["<.*>"],"rules":{"secret":{"py/object":"vakt.rules.string.StringEqualRule","val":"John"}},
                "subjects":["<.*>"],"uid":2}
                """,
                """
                {"_id":2,"actions":["<.*>"],"description":"foobar","effect":"deny",
                "resources":["<.*>"],"rules":{"secret":
                "{\\"contents\\": {\\"val\\": \\"John\\"}, \\"type\\": \\"vakt.rules.string.StringEqualRule\\"}"},
                "subjects":["<.*>"],"uid":2}
                """
            ),
            (
                """
                {"_id":3,"actions":["<.*>"],"description":"foobar","effect":"deny",
                "resources":["<.*>"],"rules":{},"subjects":["<.*>"],"uid":3}
                """,
                """
                {"_id":3,"actions":["<.*>"],"description":"foobar","effect":"deny",
                "resources":["<.*>"],"rules":{},"subjects":["<.*>"],"uid":3}
                """
            ),
            (
                """
                {"_id":4,"actions":[],"description":null,"effect":"deny","resources":[],
                "rules":{"digit":{"py/object":"test_storage_mongo.WithObject",
                "val":{"pattern":"\\\\d+","py/object":"_sre.SRE_Pattern"}},
                "num":{"py/object":"storage.test_mongo_migration.Simple","val":"123"}},"subjects":[],"uid":4}
                """,
                None
            ),
            (
                """
                {"_id":5,"actions":[],"description":null,"effect":"deny","resources":[],"rules":
                {"num":{"py/object":"storage.test_mongo_migration.Simple","val":"46"}},"subjects":[],"uid":5}
                """,
                """
                {"_id":5,"actions":[],"description":null,"effect":"deny","resources":[],"rules":
                {"num":
                "{\\"contents\\": {\\"val\\": \\"46\\"}, \\"type\\": \\"storage.test_mongo_migration.Simple\\"}"},
                "subjects":[],"uid":5}
                """
            ),
            (
                """
                {"_id":"6","uid":"6"}
                """,
                None
            ),
        ]
        for (doc, _) in docs:
            d = b_json.loads(doc)
            migration.storage.collection.insert_one(d)

        # set logger for capturing output
        l = logging.getLogger('vakt.storage.mongo')
        log_handler = self.MockLoggingHandler()
        l.setLevel(logging.INFO)
        l.addHandler(log_handler)

        migration.down()

        # test no new docs were added and no docs deleted
        assert len(docs) == len(list(migration.storage.collection.find({})))
        # test Policy.from_json() is called without errors for each doc (implicitly)
        assert len(docs) == len(list(migration.storage.get_all(1000, 0)))
        # test string contents of each doc
        for (doc, expected_doc) in docs:
            if not expected_doc:  # some policies should be left as-is if not converted
                expected_doc = doc
            new_doc = migration.storage.collection.find_one({'uid': json.loads(doc)['uid']})
            assertions.assertDictEqual(json.loads(expected_doc), new_doc)
        # test failed policies report
        # info
        assert 'info' in log_handler.messages
        assert 9 == len(log_handler.messages['info'])
        assert 'Trying to migrate Policy with UID: 5' in log_handler.messages['info']
        assert 'Policy with UID: 5 was migrated' in log_handler.messages['info']
        # warn
        assert 'warning' in log_handler.messages
        assert 2 == len(log_handler.messages['warning'])
        assert "Irreversible Policy. vakt.rules.string.RegexMatchRule could not be stored in v1.1.0. Mongo doc:" in \
               log_handler.messages['warning'][0]
        assert "'_id': 1" in log_handler.messages['warning'][0]
        assert "Irreversible Policy. Custom rule class contains non-primitive data {'" in \
               log_handler.messages['warning'][1]
        assert "'_id': 4" in log_handler.messages['warning'][1]
        # error
        assert 'error' in log_handler.messages
        assert 2 == len(log_handler.messages['warning'])
        assert "Unexpected exception occurred while migrating Policy:" in log_handler.messages['error'][0]
        assert "'uid': '6'" in log_handler.messages['error'][0]
        assert 'Migration was unable to convert some Policies, but' in log_handler.messages['error'][1]
        assert 'Mongo IDs of failed Policies are:' in log_handler.messages['error'][1]
        assert "[1, 4, '6']" in log_handler.messages['error'][1]


@pytest.mark.integration
@pytest.mark.skipif(version_info() >= (1, 3, 0), reason='migration is for versions prior to 1.3.0')
class TestMigration1x1x1To1x2x0:

    class MockLoggingHandler(logging.Handler):
        def __init__(self, *args, **kwargs):
            self.messages = {}
            super().__init__(*args, **kwargs)

        def emit(self, record):
            level = record.levelname.lower()
            if level not in self.messages:
                self.messages[level] = []
            self.messages[level].append(record.getMessage())

    @pytest.fixture()
    def storage(self):
        client = create_client()
        storage = MongoStorage(client, DB_NAME, collection=COLLECTION)
        yield storage
        client[DB_NAME][COLLECTION].delete_many({})
        client.close()

    def test_order(self, storage):
        migration = Migration1x1x1To1x2x0(storage)
        assert 3 == migration.order

    def test_up(self, storage):
        migration = Migration1x1x1To1x2x0(storage)
        # prepare docs that might have been saved by users in v 1.1.1
        docs = [
            (
                """
                { "_id" : 10, "actions" : [ ], "description" : null, "effect" : "allow", "resources" : [ ], 
                "rules" : { "name" : {"py/object": "vakt.rules.string.StringEqualRule", "val": "Max" },
                "secret" : {"py/object": "vakt.rules.string.StringEqualRule", "val": "i-am-a-foo"} }, 
                "subjects" : [ ], "uid" : 10 }
                """,
                """
                { "_id" : 10, "actions" : [ ], 
                "context" : { "name" : {"py/object": "vakt.rules.string.Equal", "val": "Max" },
                "secret" : {"py/object": "vakt.rules.string.Equal", "val": "i-am-a-foo"} }, 
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ ], "type": 1, "uid" : 10 }
                """
            ),
            (
                """
                { "_id" : 20, "actions" : [ "<.*>" ], "description" : "foo bar", "effect" : "allow",
                "resources" : [ "<.*>" ], "rules" : { "secret" :
                { "py/object": "vakt.rules.string.StringEqualRule", "val": "John"} }, "subjects" : [ "<.*>" ],
                "uid" : 20 }
                """,
                """
                { "_id" : 20, "actions" : [ "<.*>" ], "context" : { "secret" :
                { "py/object": "vakt.rules.string.Equal", "val": "John"} },
                "description" : "foo bar", "effect" : "allow",
                "resources" : [ "<.*>" ], "subjects" : [ "<.*>" ], "type": 1,
                "uid" : 20 }
                """
            ),
            (
                """
                { "_id" : 30, "actions" : [ "<.*>" ], "description" : "foo bar", "effect" : "allow",
                "resources" : [ "<.*>" ], "rules" : {  }, "subjects" : [ "<.*>" ], "uid" : 30 }
                """,
                """
                { "_id" : 30, "actions" : [ "<.*>" ], "context" : {  }, "description" : "foo bar", "effect" : "allow",
                "resources" : [ "<.*>" ], "subjects" : [ "<.*>" ], "type": 1, "uid" : 30 }
                """
            ),
            (
                """
                { "_id" : 40, "actions" : [ "<.*>" ], "description" : null, "effect" : "allow",  "resources" : ["<.*>"],
                "rules" : { "a" : {"py/object": "vakt.rules.string.StringEqualRule", "val": "foo"},
                "num" : { "py/object": "storage.test_mongo_migration.Simple", "val": "123"} }, "subjects" : ["<.*>"],
                "uid" : 40 }
                """,
                """
                { "_id" : 40, "actions" : [ "<.*>" ], "context" : { "a" : 
                {"py/object": "vakt.rules.string.Equal", "val": "foo"},
                "num" : { "py/object": "storage.test_mongo_migration.Simple", "val": "123"} }, "description" : null, 
                "effect" : "allow",  "resources" : ["<.*>"],
                 "subjects" : ["<.*>"], "type": 1,
                "uid" : 40 }
                """
            ),
            (
                """
                { "_id" : 50, "actions" : [ ], "description" : null, "effect" : "allow", "resources" : [ ],
                "rules" : { "num" : { "py/object": "storage.test_mongo_migration.Simple", "val": "46"} },
                "subjects" : [ ], "uid" : 50 }
                """,
                """
                { "_id" : 50, "actions" : [ ], 
                "context" : { "num" : { "py/object": "storage.test_mongo_migration.Simple", "val": "46"} },
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ ], "type": 1, "uid" : 50 }
                """
            ),
        ]
        for (doc, _) in docs:
            d = b_json.loads(doc)
            storage.collection.insert_one(d)

        # set logger for capturing output
        l = logging.getLogger('vakt.storage.mongo')
        log_handler = self.MockLoggingHandler()
        l.setLevel(logging.INFO)
        l.addHandler(log_handler)

        migration.up()

        # test indices exist
        created_indices = [i['name'] for i in storage.collection.list_indexes()]
        assert created_indices == ['_id_', 'type_idx']

        # test no new docs were added and no docs deleted
        assert len(docs) == len(list(storage.collection.find({})))
        # test Policy.from_json() is called without errors for each doc (implicitly)
        assert len(docs) == len(list(storage.get_all(1000, 0)))
        # test string contents of each doc
        for (doc, result_doc) in docs:
            new_doc = storage.collection.find_one({'uid': json.loads(doc)['uid']})
            expected = result_doc.replace("\n", '').replace(' ', '')
            actual = json.dumps(new_doc, sort_keys=True).replace("\n", '').replace(' ', '')
            assert expected == actual
        # test we can retrieve policy for inquiry
        inq = Inquiry(action='foo', resource='bar', subject='Max', context={'val': 'foo', 'num': '123'})
        assert len(docs) == len(list(storage.find_for_inquiry(inq, RegexChecker())))
        # test failed policies report
        assert 10 == len(log_handler.messages.get('info', []))
        assert 0 == len(log_handler.messages.get('warning', []))
        assert 0 == len(log_handler.messages.get('error', []))

    def test_down(self, storage):
        assertions = unittest.TestCase('__init__')
        migration = Migration1x1x1To1x2x0(storage)
        # prepare docs that might have been saved by users in v 1.2.0
        docs = [
            (
                """
                { "_id" : 10, "actions" : [ ], 
                "context" : { "name" : {"py/object": "vakt.rules.string.Equal", "val": "Max" },
                "secret" : {"py/object": "vakt.rules.string.StringEqualRule", "val": "i-am-a-foo"} }, 
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ ], "type": 1, "uid" : 10 }
                """,
                """
                { "_id" : 10, "actions" : [ ], "description" : null, "effect" : "allow", "resources" : [ ], 
                "rules" : { "name" : {"py/object": "vakt.rules.string.StringEqualRule", "val": "Max" },
                "secret" : {"py/object": "vakt.rules.string.StringEqualRule", "val": "i-am-a-foo"} }, 
                "subjects" : [ ], "uid" : 10 }
                """,
            ),
            (
                """
                { "_id" : 20, "actions" : [ ], 
                "context" : { },
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ ], "type": 2, "uid" : 20 }
                """,
                None,
            ),
            (
                """
                { "_id" : 30, "actions" : [ ], 
                "context" : { "name" : {"py/object": "vakt.rules.operator.Eq", "val": "Max" },
                "secret" : {"py/object": "vakt.rules.string.StringEqualRule", "val": "i-am-a-foo"} }, 
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ ], "type": 1, "uid" : 30 }
                """,
                None,
            ),
            (
                """
                { "_id" : 40, "actions" : [ ], 
                "context" : { "name" : {"py/object": "vakt.rules.string.StartsWith", "val": "Max" } }, 
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ ], "type": 1, "uid" : 40 }
                """,
                None,
            ),
        ]
        for (doc, _) in docs:
            d = b_json.loads(doc)
            migration.storage.collection.insert_one(d)

        # set logger for capturing output
        l = logging.getLogger('vakt.storage.mongo')
        log_handler = self.MockLoggingHandler()
        l.setLevel(logging.INFO)
        l.addHandler(log_handler)

        migration.down()

        # test index on 'type' was deleted
        assert ['_id_'] == [i['name'] for i in storage.collection.list_indexes()]

        # test no new docs were added and no docs deleted
        assert len(docs) == len(list(storage.collection.find({})))
        # test Policy.from_json() is called without errors for each doc (implicitly)
        assert len(docs) == len(list(storage.get_all(1000, 0)))
        # test string contents of each doc
        for (doc, expected_doc) in docs:
            if not expected_doc:  # some policies should be left as-is if not converted
                expected_doc = doc
            new_doc = storage.collection.find_one({'uid': json.loads(doc)['uid']})
            assertions.assertDictEqual(json.loads(expected_doc), new_doc)
        # test failed policies report
        # info
        assert 'info' in log_handler.messages
        assert 5 == len(log_handler.messages.get('info', []))
        assert 'Trying to migrate Policy with UID: 10' in log_handler.messages['info']
        assert 'Trying to migrate Policy with UID: 40' in log_handler.messages['info']
        assert 'Policy with UID: 10 was migrated' in log_handler.messages['info']
        # warn
        assert 'warning' in log_handler.messages
        assert 3 == len(log_handler.messages['warning'])
        assert "Policy is not of a string-based type, so not supported in < v1.2.0. Mongo doc:" in \
               log_handler.messages['warning'][0]
        assert "'_id': 20" in log_handler.messages['warning'][0]
        assert "Irreversible Policy. Context contains rule that exist only in >= v1.2.0: {'" in \
               log_handler.messages['warning'][1]
        assert "'uid': 30" in log_handler.messages['warning'][1]
        assert "Irreversible Policy. Context contains rule that exist only in >= v1.2.0: {'" in \
               log_handler.messages['warning'][1]
        assert "'uid': 40" in log_handler.messages['warning'][2]
        # error
        assert 'error' in log_handler.messages
        assert 1 == len(log_handler.messages['error'])
        assert 'Migration was unable to convert some Policies, but' in log_handler.messages['error'][0]
        assert 'Mongo IDs of failed Policies are:' in log_handler.messages['error'][0]
        assert "[20, 30, 40]" in log_handler.messages['error'][0]


@pytest.mark.integration
class TestMigration1x2x0To1x4x0:
    @pytest.fixture()
    def storage(self):
        client = create_client()
        storage = MongoStorage(client, DB_NAME, collection=COLLECTION)
        yield storage
        client[DB_NAME][COLLECTION].delete_many({})
        client.close()

    def test_order(self, storage):
        migration = Migration1x2x0To1x4x0(storage)
        assert 4 == migration.order

    def test_up(self, storage):
        migration = Migration1x2x0To1x4x0(storage)
        # prepare docs that might have been saved by users in v 1.2.0
        docs = [
            (
                """
                { "_id" : 10, "actions" : [ ], 
                "context" : { "name" : {"py/object": "vakt.rules.string.Equal", "val": "Max" },
                "secret" : {"py/object": "vakt.rules.string.Equal", "val": "i-am-a-foo"} }, 
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ ], "type": 1, "uid" : 10 }
                """,
                """
                { "_id" : 10, "actions" : [ ], "actions_compiled_regex" : [ ],
                "context" : { "name" : {"py/object": "vakt.rules.string.Equal", "val": "Max" },
                "secret" : {"py/object": "vakt.rules.string.Equal", "val": "i-am-a-foo"} }, 
                "description" : null, "effect" : "allow", "resources" : [ ],
                "resources_compiled_regex" : [ ],
                "subjects" : [ ], "subjects_compiled_regex" : [ ], "type": 1, "uid" : 10 }
                """
            ),
            (
                """
                { "_id" : 20, "actions" : [ "<.*>" ], "context" : { "secret" :
                { "py/object": "vakt.rules.string.Equal", "val": "John"} },
                "description" : "foo bar", "effect" : "allow",
                "resources" : [ "<.*>" ], "subjects" : [ "<.*>" ], "type": 1,
                "uid" : 20 }
                """,
                """
                { "_id" : 20,
                "actions" : [ "<.*>" ], "actions_compiled_regex" : [ "^(.*)$" ],
                "context" : { "secret" : { "py/object": "vakt.rules.string.Equal", "val": "John"} },
                "description" : "foo bar", "effect" : "allow",
                "resources" : [ "<.*>" ], "resources_compiled_regex" : [ "^(.*)$" ],
                "subjects" : [ "<.*>" ], "subjects_compiled_regex" : [ "^(.*)$" ],
                "type": 1, "uid" : 20 }
                """
            ),
            (
                """
                { "_id" : 30, "actions" : [ "get", "list" ],
                "context" : {  }, "description" : "foo bar",
                "effect" : "allow",
                "resources" : [ "fax", "<[pP]rinter>" ], "subjects" : [ "<.*>" ], "type": 1, "uid" : 30 }
                """,
                """
                { "_id" : 30, "actions" : [ "get", "list" ], "actions_compiled_regex" : [ "get", "list" ],
                "context" : {  }, "description" : "foo bar",
                "effect" : "allow",
                "resources" : [ "fax", "<[pP]rinter>" ], "resources_compiled_regex" : [ "fax", "^([pP]rinter)$" ],
                "subjects" : [ "<.*>" ], "subjects_compiled_regex" : [ "^(.*)$" ],
                "type": 1, "uid" : 30 }
                """
            ),
            (
                """
                { "_id" : 40, "actions" : [ ], "context" : { },
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ { "name" : {"py/object": "vakt.rules.string.StartsWith", "val": "Max" } } ],
                "type": 2, "uid" : 40 }
                """,
                """
                { "_id" : 40, "actions" : [ ], "context" : { },
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ { "name" : {"py/object": "vakt.rules.string.StartsWith", "val": "Max" } } ],
                "type": 2, "uid" : 40 }
                """
            ),
        ]
        for (doc, _) in docs:
            d = b_json.loads(doc)
            storage.collection.insert_one(d)

        migration.up()

        # test indices exist
        created_indices = [i['name'] for i in storage.collection.list_indexes()]
        assert created_indices == [
            '_id_',
            'actions_compiled_regex_idx',
            'subjects_compiled_regex_idx',
            'resources_compiled_regex_idx'
        ]

        # test no new docs were added and no docs deleted
        assert len(docs) == len(list(storage.collection.find({})))
        # test Policy.from_json() is called without errors for each doc (implicitly)
        assert len(docs) == len(list(storage.get_all(1000, 0)))
        # test string contents of each doc
        for (doc, result_doc) in docs:
            new_doc = storage.collection.find_one({'uid': json.loads(doc)['uid']})
            expected = result_doc.replace("\n", '').replace(' ', '')
            actual = json.dumps(new_doc, sort_keys=True).replace("\n", '').replace(' ', '')
            assert expected == actual
        # test we can retrieve policy for inquiry: 2 policies fit in our case
        inq = Inquiry(action='get', resource='printer', subject='Max')
        assert 2 == len(list(storage.find_for_inquiry(inq, RegexChecker())))

    def test_down(self, storage):
        assertions = unittest.TestCase('__init__')
        migration = Migration1x2x0To1x4x0(storage)
        # prepare docs that might have been saved by users in v 1.4.0
        docs = [
            (
                """
                { "_id" : 10, "actions" : [ ], "actions_compiled_regex" : [ ],
                "context" : { "name" : {"py/object": "vakt.rules.string.Equal", "val": "Max" },
                "secret" : {"py/object": "vakt.rules.string.Equal", "val": "i-am-a-foo"} }, 
                "description" : null, "effect" : "allow", "resources" : [ ],
                "resources_compiled_regex" : [ ],
                "subjects" : [ ], "subjects_compiled_regex" : [ ], "type": 1, "uid" : 10 }
                """,
                """
                { "_id" : 10, "actions" : [ ], 
                "context" : { "name" : {"py/object": "vakt.rules.string.Equal", "val": "Max" },
                "secret" : {"py/object": "vakt.rules.string.Equal", "val": "i-am-a-foo"} }, 
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ ], "type": 1, "uid" : 10 }
                """
            ),
            (
                """
                { "_id" : 20,
                "actions" : [ "<.*>" ], "actions_compiled_regex" : [ "^(.*)$" ],
                "context" : { "secret" : { "py/object": "vakt.rules.string.Equal", "val": "John"} },
                "description" : "foo bar", "effect" : "allow",
                "resources" : [ "<.*>" ], "resources_compiled_regex" : [ "^(.*)$" ],
                "subjects" : [ "<.*>" ], "subjects_compiled_regex" : [ "^(.*)$" ],
                "type": 1, "uid" : 20 }
                """,
                """
                { "_id" : 20, "actions" : [ "<.*>" ], "context" : { "secret" :
                { "py/object": "vakt.rules.string.Equal", "val": "John"} },
                "description" : "foo bar", "effect" : "allow",
                "resources" : [ "<.*>" ], "subjects" : [ "<.*>" ], "type": 1,
                "uid" : 20 }
                """
            ),
            (
                """
                { "_id" : 30, "actions" : [ "get", "list" ], "actions_compiled_regex" : [ "get", "list" ],
                "context" : {  }, "description" : "foo bar",
                "effect" : "allow",
                "resources" : [ "fax", "<[pP]rinter>" ], "resources_compiled_regex" : [ "fax", "^([pP]rinter)$" ],
                "subjects" : [ "<.*>" ], "subjects_compiled_regex" : [ "^(.*)$" ],
                "type": 1, "uid" : 30 }
                """,
                """
                { "_id" : 30, "actions" : [ "get", "list" ],
                "context" : {  }, "description" : "foo bar",
                "effect" : "allow",
                "resources" : [ "fax", "<[pP]rinter>" ], "subjects" : [ "<.*>" ], "type": 1, "uid" : 30 }
                """
            ),
            (
                """
                { "_id" : 40, "actions" : [ ], "context" : { },
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ { "name" : {"py/object": "vakt.rules.string.StartsWith", "val": "Max" } } ],
                "type": 2, "uid" : 40 }
                """,
                """
                { "_id" : 40, "actions" : [ ], "context" : { },
                "description" : null, "effect" : "allow", "resources" : [ ],
                "subjects" : [ { "name" : {"py/object": "vakt.rules.string.StartsWith", "val": "Max" } } ],
                "type": 2, "uid" : 40 }
                """
            ),
        ]
        for (doc, _) in docs:
            d = b_json.loads(doc)
            migration.storage.collection.insert_one(d)

        migration.down()

        # test indices on *_compiled_regex were deleted
        assert ['_id_'] == [i['name'] for i in storage.collection.list_indexes()]

        # test no new docs were added and no docs deleted
        assert len(docs) == len(list(storage.collection.find({})))
        # test Policy.from_json() is called without errors for each doc (implicitly)
        assert len(docs) == len(list(storage.get_all(1000, 0)))
        # test string contents of each doc
        for (doc, expected_doc) in docs:
            new_doc = storage.collection.find_one({'uid': json.loads(doc)['uid']})
            assertions.assertDictEqual(json.loads(expected_doc), new_doc)



from vakt.storage.memory import MemoryStorage
from vakt.util import Observer


# todo - move all helper and unit-test example classes and functions here

class MemoryStorageYieldingExample(MemoryStorage):
    def get_all(self, limit, offset):
        self._check_limit_and_offset(limit, offset)
        result = [v for v in self.policies.values()]
        if offset > len(result) or limit == 0:
            return []
        for p in result[offset:limit+offset]:
            yield p


class MemoryStorageYieldingExample2(MemoryStorageYieldingExample):
    def find_for_inquiry(self, inquiry, checker=None):
        for p in super().find_for_inquiry(inquiry, checker):
            yield p


class CountObserver(Observer):
    def __init__(self):
        self.count = 0

    def update(self):
        self.count += 1



from abc import abstractmethod
import pytest

from vakt.storage.migration import Migration, MigrationSet, Migrator


# SETUP

class MMigration(Migration):
    def __init__(self, db):
        self.db = db

    @property
    @abstractmethod
    def letter(self):
        pass

    def up(self):
        self.db['data'] += self.letter

    def down(self):
        self.db['data'] = self.db['data'].replace(self.letter, '')


class M1(MMigration):
    @property
    def order(self):
        return 1

    @property
    def letter(self):
        return 'A'


class M2(MMigration):
    @property
    def order(self):
        return 2

    @property
    def letter(self):
        return 'B'


class M3(MMigration):
    @property
    def order(self):
        return 3

    @property
    def letter(self):
        return 'C'


class MigrationsForString(MigrationSet):
    def __init__(self, db):
        self.db = db

    def migrations(self):
        return [
            M1(self.db),
            M2(self.db),
            M3(self.db),
        ]

    def save_applied_number(self, number):
        self.db['ver'] = number

    def last_applied(self):
        return self.db['ver']


# TESTS
@pytest.mark.parametrize('db, expect_data, expect_saved_version', [
    ({'data': '', 'ver': 0}, 'ABC', 3),
    ({'data': 'Foo', 'ver': 0}, 'FooABC', 3),
    ({'data': '', 'ver': 2}, 'C', 3),
    ({'data': '', 'ver': 3}, '', 3),
])
def test_migrator_up(db, expect_data, expect_saved_version):
    m = Migrator(MigrationsForString(db))
    m.up()
    assert expect_data == db['data']
    assert expect_saved_version == db['ver']


@pytest.mark.parametrize('db, expect_data, expect_saved_version, number', [
    ({'data': '', 'ver': 0}, 'A', 1, 1),
    ({'data': '', 'ver': 0}, 'C', 3, 3),
    ({'data': '', 'ver': 0}, '', 0, 5),
    ({'data': 'AB', 'ver': 2}, 'AB', 2, 1),
    ({'data': 'AB', 'ver': 2}, 'AB', 2, 2),
    ({'data': 'AB', 'ver': 2}, 'ABC', 3, 3),
    ({'data': 'ABC', 'ver': 3}, 'ABC', 3, 5),
])
def test_migrator_up_with_number(db, expect_data, expect_saved_version, number):
    m = Migrator(MigrationsForString(db))
    m.up(number)
    assert expect_data == db['data']
    assert expect_saved_version == db['ver']


@pytest.mark.parametrize('db, expect_data, expect_saved_version', [
    ({'data': 'ABC', 'ver': 3}, '', 0),
    ({'data': 'FooABC', 'ver': 3}, 'Foo', 0),
    ({'data': 'AB', 'ver': 2}, '', 0),
    ({'data': '', 'ver': 3}, '', 0),
])
def test_migrator_down(db, expect_data, expect_saved_version):
    m = Migrator(MigrationsForString(db))
    m.down()
    assert expect_data == db['data']
    assert expect_saved_version == db['ver']


@pytest.mark.parametrize('db, expect_data, expect_saved_version, number', [
    ({'data': 'A', 'ver': 1}, '', 0, 1),
    ({'data': 'ABC', 'ver': 3}, 'AB', 2, 3),
    ({'data': 'ABC', 'ver': 3}, 'ABC', 3, 5),
    ({'data': 'ABC', 'ver': 3}, 'AC', 1, 2),
    ({'data': 'ABC', 'ver': 3}, 'BC', 0, 1),
    ({'data': 'AB', 'ver': 2}, 'AB', 2, 3),
    ({'data': 'AB', 'ver': 2}, 'A', 1, 2),
    ({'data': 'AB', 'ver': 2}, 'B', 0, 1),
])
def test_migrator_down_with_number(db, expect_data, expect_saved_version, number):
    m = Migrator(MigrationsForString(db))
    m.down(number)
    assert expect_data == db['data']
    assert expect_saved_version == db['ver']



import pytest

from vakt.exceptions import PolicyExistsError, UnknownCheckerType
from vakt.checker import RegexChecker


class MyChecker(RegexChecker):
    pass


@pytest.mark.parametrize('uid, message', [
    ('', 'Conflicting UID = '),
    ('123', 'Conflicting UID = 123'),
])
def test_poicy_exists_message(uid, message):
    ex = PolicyExistsError(uid)
    assert message == str(ex)


@pytest.mark.parametrize('obj, message', [
    (RegexChecker(), "Can't determine Checker type. Given: RegexChecker"),
    (MyChecker(), "Can't determine Checker type. Given: MyChecker"),
    (1, "Can't determine Checker type. Given: int"),
])
def test_unknown_checker_message(obj, message):
    ex = UnknownCheckerType(obj)
    assert message == str(ex)



from vakt.storage.memory import MemoryStorage
from vakt import Policy, Inquiry, RulesChecker, ALLOW_ACCESS
from vakt.cache import create_cached_guard
from vakt.rules import Eq


class TestAllowanceCache:

    def test_same_inquiries_are_cached(self):
        guard, storage, cache = create_cached_guard(MemoryStorage(), RulesChecker(), maxsize=256)
        p1 = Policy(1, actions=[Eq('get')], resources=[Eq('book')], subjects=[Eq('Max')], effect=ALLOW_ACCESS)
        storage.add(p1)
        inq1 = Inquiry(action='get', resource='book', subject='Max')
        inq2 = Inquiry(action='get', resource='book', subject='Jamey')
        assert guard.is_allowed(inq1)
        assert guard.is_allowed(inq1)
        assert guard.is_allowed(inq1)
        assert guard.is_allowed(inq1)
        assert guard.is_allowed(inq1)
        assert not guard.is_allowed(inq2)
        assert 4 == cache.info().hits
        assert 2 == cache.info().misses
        assert 2 == cache.info().currsize

    def test_cache_is_invalidated_on_policy_change(self):
        def assert_after_modification():
            assert 0 == cache.info().hits
            assert 0 == cache.info().misses
            assert 0 == cache.info().currsize
            assert not guard.is_allowed(inq1)
            assert not guard.is_allowed(inq1)
            assert guard.is_allowed(inq2)
            assert guard.is_allowed(inq2)
            assert guard.is_allowed(inq2)
            assert 3 == cache.info().hits
            assert 2 == cache.info().misses
            assert 2 == cache.info().currsize

        inq1 = Inquiry(action='get', resource='book', subject='Max')
        inq2 = Inquiry(action='get', resource='book', subject='Jim')
        guard, storage, cache = create_cached_guard(MemoryStorage(), RulesChecker(), maxsize=256)
        p1 = Policy(1, actions=[Eq('get')], resources=[Eq('book')], subjects=[Eq('Max')], effect=ALLOW_ACCESS)
        p2 = Policy(2, actions=[Eq('get')], resources=[Eq('magazine')], subjects=[Eq('Max')], effect=ALLOW_ACCESS)
        storage.add(p1)
        assert guard.is_allowed(inq1)
        assert guard.is_allowed(inq1)
        assert 1 == cache.info().hits
        assert 1 == cache.info().misses
        assert 1 == cache.info().currsize
        # start modifications
        p1.subjects = [Eq('Jim')]
        storage.update(p1)
        assert_after_modification()
        storage.add(p2)
        assert_after_modification()
        storage.delete(p2)
        assert_after_modification()



import operator
import random
import types
import unittest
import uuid
from operator import attrgetter

import pytest
from sqlalchemy.orm import sessionmaker, scoped_session

from vakt.checker import StringExactChecker, StringFuzzyChecker, RegexChecker, RulesChecker
from vakt.effects import ALLOW_ACCESS, DENY_ACCESS
from vakt.exceptions import PolicyExistsError, UnknownCheckerType
from vakt.guard import Inquiry, Guard
from vakt.policy import Policy
from vakt.rules.logic import Any
from vakt.rules.operator import Eq
from vakt.rules.string import Equal
from vakt.storage.memory import MemoryStorage
from vakt.storage.sql import SQLStorage
from vakt.storage.sql.model import Base

from . import create_test_sql_engine


@pytest.mark.sql_integration
class TestSQLStorage:

    @pytest.yield_fixture
    def session(self):
        engine = create_test_sql_engine()
        Base.metadata.create_all(engine)
        session = scoped_session(sessionmaker(bind=engine))
        yield session
        Base.metadata.drop_all(engine)

    @pytest.yield_fixture
    def st(self, session):
        yield SQLStorage(scoped_session=session)
        session.remove()

    def test_add(self, st):
        id = str(uuid.uuid4())
        p = Policy(
            uid=id,
            description='foo bar баз',
            subjects=('Edward Rooney', 'Florence Sparrow'),
            actions=['<.*>'],
            resources=['<.*>'],
            context={
                'secret': Equal('i-am-a-teacher'),
            },
        )
        st.add(p)
        back = st.get(id)
        assert id == back.uid
        assert 'foo bar баз' == back.description
        assert isinstance(back.context['secret'], Equal)

        st.add(Policy('2', actions=[Eq('get'), Eq('put')], subjects=[Any()], resources=[{'books': Eq('Harry')}]))
        assert '2' == st.get('2').uid
        assert 2 == len(st.get('2').actions)
        assert 1 == len(st.get('2').subjects)
        assert isinstance(st.get('2').subjects[0], Any)
        assert 1 == len(st.get('2').resources)
        assert isinstance(st.get('2').resources[0]['books'], Eq)
        assert 'Harry' == st.get('2').resources[0]['books'].val

    def test_policy_create_existing(self, st):
        id = str(uuid.uuid4())
        st.add(Policy(id, description='foo'))
        with pytest.raises(PolicyExistsError):
            st.add(Policy(id, description='bar'))

    def test_get(self, st):
        st.add(Policy('1'))
        st.add(Policy(2, description='some text'))
        assert isinstance(st.get('1'), Policy)
        # SQL storage stores all uid as string
        assert '1' == st.get('1').uid
        assert '2' == st.get('2').uid
        assert 'some text' == st.get('2').description

    def test_get_nonexistent(self, st):
        assert None is st.get('123456789')

    @pytest.mark.parametrize('limit, offset, result', [
        (50, 0, 20),
        (11, 1, 11),
        (50, 5, 15),
        (20, 0, 20),
        (20, 1, 19),
        (19, 0, 19),
        (20, 5, 15),
        (0, 0, 0),
        (0, 10, 0),
        (1, 0, 1),
        (5, 4, 5),
    ])
    def test_get_all(self, st, limit, offset, result):
        for i in range(20):
            desc = ''.join(random.choice('abcde') for _ in range(30))
            st.add(Policy(str(i), description=desc))
        policies = list(st.get_all(limit=limit, offset=offset))
        assert result == len(policies)

    def test_get_all_check_policy_properties(self, st):
        p = Policy(
            uid='1',
            description='foo bar баз',
            subjects=('Edward Rooney', 'Florence Sparrow'),
            actions=['<.*>'],
            resources=['<.*>'],
            context={
                'secret': Equal('i-am-a-teacher'),
            },
        )
        st.add(p)
        policies = list(st.get_all(100, 0))
        assert 1 == len(policies)
        assert '1' == policies[0].uid
        assert 'foo bar баз' == policies[0].description
        assert 2 == len(policies[0].subjects)
        assert 'Edward Rooney' in policies[0].subjects
        assert 'Florence Sparrow' in policies[0].subjects
        assert ['<.*>'] == policies[0].actions
        assert ['<.*>'] == policies[0].resources
        assert isinstance(policies[0].context['secret'], Equal)

    def test_get_all_with_incorrect_args(self, st):
        for i in range(10):
            st.add(Policy(str(i), description='foo'))
        with pytest.raises(ValueError) as e:
            list(st.get_all(-1, 9))
        assert "Limit can't be negative" == str(e.value)
        with pytest.raises(ValueError) as e:
            list(st.get_all(0, -3))
        assert "Offset can't be negative" == str(e.value)

    def test_get_all_returns_generator(self, st):
        st.add(Policy('1'))
        st.add(Policy('2'))
        found = st.get_all(500, 0)
        assert isinstance(found, types.GeneratorType)
        l = []
        for p in found:
            l.append(p.uid)
        assert 2 == len(l)

    def test_get_all_ascending_sorting_order(self, st):
        for i in range(1, 20):
            st.add(Policy(i))
        expected_uids = sorted(list(map(str, range(1, 20))))
        assert expected_uids == list(map(attrgetter('uid'), st.get_all(30, 0)))

    @pytest.mark.parametrize('checker, expect_number', [
        (None, 6),
        (RulesChecker(), 2),
        (StringExactChecker(), 1),
        (StringFuzzyChecker(), 1),
    ])
    def test_find_for_inquiry_returns_existing_policies(self, st, checker, expect_number):
        st.add(Policy('1', subjects=['<[mM]ax>', '<.*>'], actions=['delete'], resources=['server']))
        st.add(Policy('2', subjects=['Ji<[mM]+>'], actions=['delete'], resources=[r'server<\s*>']))
        st.add(Policy('3', subjects=['sam<.*>', 'foo']))
        st.add(Policy('5', subjects=['Jim'], actions=['delete'], resources=['server']))
        st.add(Policy('4', subjects=[{'stars': Eq(90)}, Eq('Max')]))
        st.add(Policy('6', subjects=[Eq('Jim'), Eq('Nina')]))
        inquiry = Inquiry(subject='Jim', action='delete', resource='server')
        found = st.find_for_inquiry(inquiry, checker)
        found = list(found)
        assert expect_number == len(found)

    @pytest.mark.parametrize('dialect, expect_number', [
        ('sqlite', 7),
        ('mysql', 5),
        ('postgresql', 5),
    ])
    def test_regex_checker_find_for_inquiry_policies_count_return_by_storage(self, st, dialect, expect_number):
        if st.dialect != dialect:
            pytest.skip('skipping for %s dialect' % dialect)
        st.add(Policy('1', subjects=['<[mM]ax>', '<.*>'], actions=['delete'], resources=['server']))
        st.add(Policy('2', subjects=['Ji<[mM]+>'], actions=['delete'], resources=[r'server<\s*>']))
        st.add(Policy('3', subjects=['sam<.*>', 'foo']))
        st.add(Policy('5', subjects=['Jim'], actions=['delete'], resources=['server']))
        st.add(Policy('6', subjects=['Ji<[mM]+>'], actions=[r'<[a-zA-Z]{6}>'], resources=['serve<(r|rs)>']))
        st.add(Policy('7', subjects=['Ji<[mM]+>'], actions=[r'<[a-zA-Z]{6}>'], resources=['serve<(u|rs)>']))
        st.add(Policy('8', subjects=['Ji<[mM]+>'], actions=[r'<[a-zA-Z]{6}>'], resources=['serve<(u|rs)>', 'server']))
        st.add(Policy('40', subjects=[{'stars': Eq(90)}, Eq('Max')]))
        st.add(Policy('60', subjects=[Eq('Jim'), Eq('Nina')]))
        inquiry = Inquiry(subject='Jim', action='delete', resource='server')
        found = st.find_for_inquiry(inquiry, RegexChecker())
        found = list(found)
        assert expect_number == len(found)

    def test_find_for_inquiry_with_exact_string_checker(self, st):
        st.add(Policy('1', subjects=['max', 'bob'], actions=['get'], resources=['books', 'comics', 'magazines']))
        st.add(Policy('2', subjects=['maxim'], actions=['get'], resources=['books', 'comics', 'magazines']))
        st.add(Policy('3', subjects=['sam', 'nina']))
        st.add(Policy('4', subjects=[Eq('sam'), Eq('nina')]))
        inquiry = Inquiry(subject='max', action='get', resource='books')
        found = st.find_for_inquiry(inquiry, StringExactChecker())
        found = list(found)
        assert 1 == len(found)
        assert '1' == found[0].uid

    def test_find_for_inquiry_with_fuzzy_string_checker(self, st):
        st.add(Policy('1', subjects=['max', 'bob'], actions=['get'], resources=['books', 'comics', 'magazines']))
        st.add(Policy('2', subjects=['maxim'], actions=['get'], resources=['books', 'foos']))
        st.add(Policy('3', subjects=['Max'], actions=['get'], resources=['books', 'comics']))
        st.add(Policy('4', subjects=['sam', 'nina']))
        st.add(Policy('5', subjects=[Eq('sam'), Eq('nina')]))
        inquiry = Inquiry(subject='max', action='et', resource='oo')
        found = st.find_for_inquiry(inquiry, StringFuzzyChecker())
        found = list(found)
        assert 2 == len(found)
        ids = [found[0].uid, found[1].uid]
        assert '1' in ids
        assert '2' in ids
        inquiry = Inquiry(subject='Max', action='get', resource='comics')
        found = st.find_for_inquiry(inquiry, StringFuzzyChecker())
        found = list(found)
        assert 1 == len(found)
        assert '3' == found[0].uid

    @pytest.mark.parametrize('policies, inquiry, expected_reference', [
        (
            [
                Policy(
                    uid=1,
                    actions=['get', 'post'],
                    effect=ALLOW_ACCESS,
                    resources=['<.*>'],
                    subjects=['<[Mm]ax>', '<Jim>']
                ),
            ],
            Inquiry(action='get', resource='printer', subject='Max'),
            True,
        ),
        (
            [
                Policy(
                    uid=1,
                    actions=['<.*>'],
                    effect=ALLOW_ACCESS,
                    resources=['<.*>'],
                    subjects=['<.*>']
                ),
            ],
            Inquiry(action='get', resource='printer', subject='Max'),
            True,
        ),
        (
            [
                Policy(
                    uid=1,
                    actions=['<.*>'],
                    effect=ALLOW_ACCESS,
                    resources=['library:books:<.+>'],
                    subjects=['<.*>']
                ),
            ],
            Inquiry(action='get', resource='library:books:dracula', subject='Max'),
            True,
        ),
        (
            [
                Policy(
                    uid=1,
                    actions=[r'<[0-9]+>'],
                    effect=ALLOW_ACCESS,
                    resources=[r'<[a-zA-Z]{1,3}>'],
                    subjects=[r'<[a-zA-Z0-9]{2}-[0-9]+>']
                ),
            ],
            Inquiry(action='12', resource='Pie', subject='Jo-1'),
            True,
        ),
    ])
    def test_find_for_inquiry_with_regex_checker(self, st, policies, inquiry, expected_reference):
        mem_storage = MemoryStorage()  # it returns all stored policies so we consider Guard as a reference
        for p in policies:
            st.add(p)
            mem_storage.add(p)
        reference_answer = Guard(mem_storage, RegexChecker()).is_allowed(inquiry)
        assert expected_reference == reference_answer, 'Check reference answer'
        assert reference_answer == Guard(st, RegexChecker()).is_allowed(inquiry), \
            'SQL storage should give the same answers as reference'

    def test_find_for_inquiry_with_rules_checker(self, st):
        assertions = unittest.TestCase('__init__')
        st.add(Policy(1, subjects=[{'name': Equal('Max')}], actions=[{'foo': Equal('bar')}]))
        st.add(Policy(2, subjects=[{'name': Equal('Max')}], actions=[{'foo': Equal('bar2')}]))
        st.add(Policy(3, subjects=['sam', 'nina']))
        st.add(Policy(4, actions=[r'<\d+>'], effect=ALLOW_ACCESS, resources=[r'<\w{1,3}>'], subjects=[r'<\w{2}-\d+>']))
        st.add(Policy(5, subjects=[{'name': Equal('Jim')}], actions=[{'foo': Equal('bar3')}]))
        inquiry = Inquiry(subject={'name': 'max'}, action='get', resource='books')
        found = st.find_for_inquiry(inquiry, RulesChecker())
        found = list(found)
        assert 3 == len(found)
        found_uids = list(map(operator.attrgetter('uid'), found))
        found_uids.sort()
        assertions.assertListEqual(['1', '2', '5'], found_uids)

    def test_find_for_inquiry_with_unknown_checker(self, st):
        st.add(Policy('1'))
        inquiry = Inquiry(subject='sam', action='get', resource='books')
        with pytest.raises(UnknownCheckerType):
            list(st.find_for_inquiry(inquiry, Inquiry()))

    def test_find_for_inquiry_returns_generator(self, st):
        st.add(Policy('1', subjects=['max', 'bob'], actions=['get'], resources=['comics']))
        st.add(Policy('2', subjects=['max', 'bob'], actions=['get'], resources=['comics']))
        inquiry = Inquiry(subject='max', action='get', resource='comics')
        found = st.find_for_inquiry(inquiry)
        assert isinstance(found, types.GeneratorType)
        l = []
        for p in found:
            l.append(p.uid)
        assert 2 == len(l)

    def test_update(self, st):
        # SQL storage stores all uids as string
        id = str(uuid.uuid4())
        policy = Policy(id)
        st.add(policy)
        assert id == st.get(id).uid
        assert None is st.get(id).description
        assert [] == st.get(id).actions
        policy.description = 'foo'
        policy.actions = ['a', 'b', 'c']
        st.update(policy)
        assert id == st.get(id).uid
        assert 'foo' == st.get(id).description
        assert ['a', 'b', 'c'] == st.get(id).actions
        p = Policy('2', actions=[Any()], subjects=[Eq('max'), Eq('bob')])
        st.add(p)
        assert '2' == st.get('2').uid
        p.actions = [Eq('get')]
        st.update(p)
        assert 1 == len(st.get('2').actions)
        assert 'get' == st.get('2').actions[0].val

    def test_update_non_existing_does_not_create_anything(self, st):
        id = str(uuid.uuid4())
        st.update(Policy(id, actions=['get'], description='bar'))
        assert st.get(id) is None

    def test_delete(self, st):
        policy = Policy('1')
        st.add(policy)
        assert '1' == st.get('1').uid
        st.delete('1')
        assert None is st.get('1')

    def test_delete_nonexistent(self, st):
        uid = str('non-existent-id')
        st.delete(uid)
        assert None is st.get(uid)

    @pytest.mark.parametrize('effect', [
        ALLOW_ACCESS,
        DENY_ACCESS,
    ])
    def test_effect_boolean_conversion(self, st, effect):
        policy = Policy('1', effect=effect)
        st.add(policy)
        policy_back = st.get('1')
        assert effect == policy_back.effect
        policy.effect = 'foo'
        st.update(policy)
        policy.effect = effect
        st.update(policy)
        policy_back = st.get('1')
        assert effect == policy_back.effect



from sqlalchemy import Column, Integer
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative import declarative_base

from .model import Base
from ..migration import Migration, MigrationSet

MigrationBase = declarative_base()


class MigrationModel(MigrationBase):
    """
        Migration version storage
    """

    __tablename__ = 'vakt_migrations'

    id = Column(Integer, primary_key=True)
    version = Column(Integer, nullable=False)


class SQLMigrationSet(MigrationSet):
    """
        Set of migrations for SQL Storage
    """

    def __init__(self, storage):
        self.storage = storage
        self.session = storage.session
        self._index = 1

        MigrationModel.metadata.create_all(self.storage.session.bind)

    def migrations(self):
        return [Migration0To1x3x0(self.storage)]

    def save_applied_number(self, number):
        try:
            data = self.session.query(MigrationModel).get(self._index)
            # Insert if entry not found else update
            if not data:
                data = MigrationModel(id=self._index, version=number)
                self.session.add(data)
            else:
                data.version = number
            self.session.commit()
        except SQLAlchemyError as err:
            self.session.rollback()
            raise err

    def last_applied(self):
        data = self.session.query(MigrationModel).get(self._index)
        if data:
            return data.version
        return 0


class Migration0To1x3x0(Migration):
    """
        Migration between versions 0 and 1.3.0.
        This migration is initial.
    """

    def __init__(self, storage):
        self.storage = storage

    @property
    def order(self):
        return 1

    def up(self):
        Base.metadata.create_all(self.storage.session.bind)

    def down(self):
        Base.metadata.drop_all(self.storage.session.bind)



"""
Caching mechanisms for vakt
"""

import logging
from functools import lru_cache
from abc import ABCMeta, abstractmethod

from .storage.observable import ObservableMutationStorage
from .util import Observer
from .guard import Guard


__all__ = [
    'create_cached_guard',
    'EnfoldCache',
    'AllowanceCacheBackend',
]


log = logging.getLogger(__name__)


def create_cached_guard(storage, checker, cache=None, **kwargs):
    """
    Creates Guard whose `is_allowed` method calls are cached.
    It helps to increase performance for similar Inquiries in case you have static Policies set.

    :argument
    storage - Storage, that will be used by Guard and modified for further usage
    checker - Checker implementation, that will be used by Guard
    cache - argument allows you to provide your own cache implementation that must
            implement vakt.cache.AllowanceCacheBackend. If it is None the default in-memory LRU cache will be used.
            It also accepts optional keyword arguments that will be passed to a cache.
            Currently only `maxsize` is available.
    maxsize - argument allows you to specify a maximum size of a default in-memory LRU cache, (preferably a power of 2)

    :return (storage, guard, cache)
    guard - Guard whose `is_allowed` method will be cached
    storage - must be used in the code after the function call - it's an observable wrapper of the initial Storage.
              If you are not interacting with policies through this Storage cache won't behave correctly.
    cache - AllowanceCache that can be used to obtain information about a cache state
    """
    st = ObservableMutationStorage(storage)
    guard = Guard(st, checker)
    cache = AllowanceCache(guard, cache_backend=cache, **kwargs)
    st.add_listener(cache)
    return guard, st, cache


class EnfoldCache:
    """
    Wraps all underlying storage interface calls with a cache that is represented by another storage.
    By default `populate` arg is True and thus it populates cache with all the existing policies at the startup.
    This may take a long time depending on the size of policies set n your backend storage.
    Otherwise you need to set `populate` arg is False and manually call ec.populate()
    before you start working with the cache.

    Typical (and recommended) usage is:
    storage = EnfoldCache(MongoStorage(...), cache=MemoryStorage())

    Or it might be:
    storage = EnfoldCache(MongoStorage(...), cache=MemoryStorage(), populate=False)
    ...
    storage.populate()
    """

    def __init__(self, storage, cache, populate=True):
        self.storage = storage
        self.cache = cache
        self.populate_step_size = 1000
        if populate:
            self.populate()

    def populate(self):
        for p in self.storage.retrieve_all(self.populate_step_size):
            self.cache.add(p)

    def add(self, policy):
        """
        Cache storage `add`
        """
        # we aren't catching any exceptions - just letting them pass
        res = self.storage.add(policy)
        self.cache.add(policy)
        return res

    def get(self, uid):
        """
        Cache storage `get`
        """
        policy = self.cache.get(uid)
        if policy:
            return policy
        log.warning(
            '%s cache miss for get Policy with UID=%s. Trying to get it from backend storage',
            type(self).__name__, uid
        )
        return self.storage.get(uid)

    def get_all(self, limit, offset):
        """
        Cache storage `get_all`
        """
        result = list(self.cache.get_all(limit, offset))
        if len(result) > 0:
            return result
        return self.storage.get_all(limit, offset)

    def retrieve_all(self, *args, **kwargs):
        """
        Cache storage `retrieve_all`
        """
        result = list(self.cache.retrieve_all(*args, **kwargs))
        if len(result) > 0:
            return result
        return self.storage.retrieve_all(*args, **kwargs)

    def find_for_inquiry(self, inquiry, checker=None):
        """
        Cache storage `find_for_inquiry`
        """
        result = list(self.cache.find_for_inquiry(inquiry, checker))
        if len(result) > 0:
            return result
        log.warning('%s cache miss for find_for_inquiry. Trying it from backend storage', type(self).__name__)
        return self.storage.find_for_inquiry(inquiry, checker)

    def update(self, policy):
        """
        Cache storage `update`
        """
        res = self.storage.update(policy)
        self.cache.update(policy)
        return res

    def delete(self, uid):
        """
        Cache storage `delete`
        """
        res = self.storage.delete(uid)
        self.cache.delete(uid)
        return res


class AllowanceCache(Observer):
    """
    Caches hits of `is_allowed` (technically, its more tiny part: `is_allowed_no_audit`) for a given Inquiry.
    In case of a cache hit returns the cached boolean result, in case of a cache miss goes to a Storage and
    memorizes its result for future calls with the same Inquiry.
    If underlying Storage notifies it that policy-set was anyhow changed, invalidates all the cached results.

    You need to pass proper options in order to create cache of a desired type.
    Available options are:
    maxsize - maximum size of a cache
    backend - which backend will be used for caching
    type - type of a caching algorithm to be used
    """
    def __init__(self, guard, cache_backend=None, **kwargs):
        self.options = kwargs
        if cache_backend is None:
            self.cache = LRUCache(maxsize=self.options['maxsize'])
        guard.is_allowed_no_audit = self.cache.wrap(guard.is_allowed_no_audit)

    def update(self):
        """
        Is a callback for fire events on Storage modify actions.
        We need to invalidate cache since policy set is changed with each add/delete/update storage action,
        thus old Guard answers on inquiries are will be no longer valid.
        """
        self.cache.invalidate()

    def info(self):
        """
        Get information about current cache.
        """
        return self.cache.info()


class AllowanceCacheBackend(metaclass=ABCMeta):
    """
    Interface for backed cache implementations for AllowanceCache.
    All implementations should subclass it.
    """
    @abstractmethod
    def wrap(self, func):
        """
        Wrap function into its cached version and return it.
        """
        pass

    @abstractmethod
    def invalidate(self):
        """
        Invalidate the cache
        """
        pass

    @abstractmethod
    def info(self):
        """
        Get cache information.
        Preferably it should be an object with attributes:
        - hits - number of cache hits,
        - misses - number of cache misses,
        - maxsize - maximum number of elements the cache can contain,
        - currsize - number of elements the cache contains at the moment
        - ... some other useful attributes
        """
        pass


class LRUCache(AllowanceCacheBackend):
    """
    Std lib LRU in-memory cache.
    """
    def __init__(self, **kwargs):
        self.maxsize = kwargs['maxsize']
        self._wrapped_func = None

    def wrap(self, func):
        self._wrapped_func = lru_cache(self.maxsize)(func)
        return self._wrapped_func

    def invalidate(self):
        self._wrapped_func.cache_clear()

    def info(self):
        return self._wrapped_func.cache_info()



"""
All Rules that work with strings.
"""

import re
import logging
import warnings
from abc import ABCMeta

from ..rules.base import Rule


log = logging.getLogger(__name__)


__all__ = [
    'Equal',
    'PairsEqual',
    'RegexMatch',
    'StartsWith',
    'EndsWith',
    'Contains'
]


class StringRule(Rule, metaclass=ABCMeta):
    """
    Basic Rule for strings
    """
    def __init__(self, val, ci=False):
        if not isinstance(val, str):
            log.error('%s creation. Initial property should be a string', type(self).__name__)
            raise TypeError('Initial property should be a string')
        self.val = val
        self.ci = ci


class Equal(StringRule):
    """
    Rule that is satisfied if the string value equals the specified property of this rule.
    Performs case-sensitive and case-sensitive comparisons (based on `ci` (case_insensitive) flag).
    For example: context={'country': Equal('Mozambique', True)}
    """
    def satisfied(self, what, inquiry=None):
        if isinstance(what, str):
            if self.ci:
                return what.lower() == self.val.lower()
            return what == self.val
        return False


class PairsEqual(Rule):
    """
    Rule that is satisfied when given data is an array of pairs and
    those pairs are represented by equal to each other strings.
    For example: context={'scores': PairsEqual()}
    """
    def satisfied(self, what, inquiry=None):
        if not isinstance(what, list):
            return False
        for pair in what:
            if len(pair) != 2:
                return False
            if not isinstance(pair[0], str) and not isinstance(pair[1], str):
                return False
            if pair[0] != pair[1]:
                return False
        return True


class RegexMatch(Rule):
    r"""
    Rule that is satisfied when given data matches the provided regular expression.
    Note, that you should provide syntactically valid regular-expression string.
    For example: context={'file': RegexMatch(r'\.(rb|sh|py|exe)$')}
    """
    def __init__(self, pattern):
        try:
            self.regex = re.compile(pattern)
        except Exception as e:
            log.exception('%s creation. Failed to compile regexp %s', type(self).__name__, pattern)
            raise TypeError('pattern should be a valid regexp string. Error %s' % e)

    def satisfied(self, what, inquiry=None):
        return bool(self.regex.match(str(what)))


class StartsWith(StringRule):
    """
    Rule that is satisfied when given string starts with initially provided substring.
    For example: context={'file': StartsWith('Route-', ci=True)}
    """
    def satisfied(self, what, inquiry=None):
        if isinstance(what, str):
            if self.ci:
                return what.lower().startswith(self.val.lower())
            return what.startswith(self.val)
        return False


class EndsWith(StringRule):
    """
    Rule that is satisfied when given string ends with initially provided substring.
    For example: context={'file': EndsWith('.txt')}
    """
    def satisfied(self, what, inquiry=None):
        if isinstance(what, str):
            if self.ci:
                return what.lower().endswith(self.val.lower())
            return what.endswith(self.val)
        return False


class Contains(StringRule):
    """
    Rule that is satisfied when given string contains initially provided substring.
    For example: context={'file': Contains('sun')}
    """
    def satisfied(self, what, inquiry=None):
        if isinstance(what, str):
            if self.ci:
                return self.val.lower() in what.lower()
            return self.val in what
        return False


# Classes marked for removal in next releases
class StringEqualRule(Equal):
    """Deprecated in favor of Equal"""
    def __init__(self, *args, **kwargs):
        warnings.warn('StringEqualRule will be removed in version 2.0. Use Equal', DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)


class RegexMatchRule(RegexMatch):
    """Deprecated in favor of RegexMatch"""
    def __init__(self, *args, **kwargs):
        warnings.warn('RegexMatchRule will be removed in version 2.0. Use RegexMatch', DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)


class StringPairsEqualRule(PairsEqual):
    """Deprecated in favor of PairsEqual"""
    def __init__(self, *args, **kwargs):
        warnings.warn('StringPairsEqualRule will be removed in version 2.0. Use PairsEqual',
                      DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)



"""
All Rules relevant to list-related context.

Initial list against which the search will be performed is cast to a Set to speed-up further search operations.
Consequently beware that it may be slow on data insertion for large lists.

For some of the checks where list is an input (e.g. AllIn, AllNotIn, ...)
cast of input to Set is also performed.

Various search options are available:
- is item in list?
- is not item in the list?
- are all the items in the list?
- are all the items not in the list?
- is at least one of the provided items in the list?
- is at least one of the provided items not in the list?
"""

import logging
from abc import ABCMeta

from ..rules.base import Rule

log = logging.getLogger(__name__)


__all__ = [
    'In',
    'NotIn',
    'AllIn',
    'AllNotIn',
    'AnyIn',
    'AnyNotIn'
]


class ListRule(Rule, metaclass=ABCMeta):
    """
    Generic Rule for List-related checks
    """
    def __init__(self, *args):
        self.data = set(args)


class In(ListRule):
    """
    Is item in list?
    For example: actions={'method': In('read', 'write', 'delete')}. In inquiry: action={'method': 'read'}
    """
    def satisfied(self, what, inquiry=None):
        return _one_in_list(what, self.data)


class NotIn(ListRule):
    """
    Is not item in the list?
    For example: actions=[{'method': NotIn('read', 'write', 'delete')}]. In inquiry: action={'method': 'purge'}
    """
    def satisfied(self, what, inquiry=None):
        return not _one_in_list(what, self.data)


class AllIn(ListRule):
    """
    Are all the items in the list?
    For example: actions=[{'methods': AllIn('read', 'write', 'delete')}]. In inquiry: action={'method': ['purge', 'get]}
    """
    def satisfied(self, what, inquiry=None):
        if not isinstance(what, list):
            raise TypeError('Value should be of list type')
        return _all_in_list(what, self.data)


class AllNotIn(ListRule):
    """
    Are all the items not in the list?
    For example: actions=[{'methods': AllNotIn('read', 'write')}]. In inquiry: action={'method': ['list', 'get]}
    """
    def satisfied(self, what, inquiry=None):
        if not isinstance(what, list):
            raise TypeError('Value should be of list type')
        return not _all_in_list(what, self.data)


class AnyIn(ListRule):
    """
    Are any of the items in the list?
    For example: actions=[{'methods': AnyIn('read', 'write', 'delete')}]. In inquiry: action={'method': ['list', 'get]}
    """
    def satisfied(self, what, inquiry=None):
        if not isinstance(what, list):
            raise TypeError('Value should be of list type')
        return _any_in_list(what, self.data)


class AnyNotIn(ListRule):
    """
    Are any of the items not in the list?
    For example: actions=[{'methods': AnyNotIn('read', 'write')}]. In inquiry: action={'method': ['list', 'get]}
    """
    def satisfied(self, what, inquiry=None):
        if not isinstance(what, list):
            raise TypeError('Value should be of list type')
        return bool(set(what).difference(self.data))


def _one_in_list(what, data):
    return what in data


def _all_in_list(what, data):
    return set(what).issubset(data)


def _any_in_list(what, data):
    return bool(data.intersection(set(what)))



"""
All Rules that are related to comparison operators:
==, !=, <, >, <=, >=
They behave the same as you might expect from Python comparison operators.
"""

import logging
from abc import ABCMeta

from ..rules.base import Rule


log = logging.getLogger(__name__)


__all__ = [
    'Eq',
    'NotEq',
    'Greater',
    'Less',
    'GreaterOrEqual',
    'LessOrEqual'
]


class OperatorRule(Rule, metaclass=ABCMeta):
    """
    Base class for all Logic Operator Rules
    """
    def __init__(self, val):
        self.val = val


class Eq(OperatorRule):
    """
    Rule that is satisfied when two values are equal '=='.
    For example: context={'referralCount': Eq(90)}
    """
    def satisfied(self, what, inquiry=None):
        if isinstance(self.val, tuple):
            val = list(self.val)
        else:
            val = self.val
        return val == what


class NotEq(OperatorRule):
    """
    Rule that is satisfied when two values are not equal '!='.
    For example: subjects=[{'stars': NotEq(100)}]
    """
    def satisfied(self, what, inquiry=None):
        if isinstance(self.val, tuple):
            val = list(self.val)
        else:
            val = self.val
        return val != what


class Greater(OperatorRule):
    """
    Rule that is satisfied when 'what' is greater '>' than initial value.
    For example: subjects=[{'stars': Greater(100)}]
    """
    def satisfied(self, what, inquiry=None):
        return what > self.val


class Less(OperatorRule):
    """
    Rule that is satisfied when 'what' is less '<' than initial value.
    For example: subjects=[{'stars': Less(100)}]
    """
    def satisfied(self, what, inquiry=None):
        return what < self.val


class GreaterOrEqual(OperatorRule):
    """
    Rule that is satisfied when 'what' is greater or equal '>=' than initial value.
    For example: subjects=[{'stars': GreaterOrEqual(100), 'name': Eq('Jason')}]
    """
    def satisfied(self, what, inquiry=None):
        return what >= self.val


class LessOrEqual(OperatorRule):
    """
    Rule that is satisfied when 'what' is less or equal '<=' than initial value.
    For example: subjects=[{'stars': LessOrEqual(100), 'name': Eq('Jason')}]
    """
    def satisfied(self, what, inquiry=None):
        return what <= self.val



import os

import pytest
from sqlalchemy import create_engine, text
from sqlalchemy.exc import OperationalError
from sqlalchemy.event import listens_for
from sqlalchemy.pool import Pool

from vakt.storage.sql.model import Base


# Need for switching on case sensitive LIKE statements on Sqlite
@listens_for(Pool, 'connect')
def run_on_connect(dbapi_con, connection_record):
    try:
        dbapi_con.execute('pragma case_sensitive_like=ON')
    except:
        pass


def create_test_sql_engine():
    dsn = os.getenv('DATABASE_DSN')
    if not dsn:
        pytest.exit('Please set DATABASE_DSN env variable with the target database DSN, ex: sqlite:///:memory:')
    engine = create_engine(dsn, encoding='utf-8')
    try:
        engine.execute(text('select 1'))
    except OperationalError as e:
        pytest.exit('DATABASE_DSN is not correct. Error: %s' % e)
    return engine



"""
Vakt is an SDK for access policies.
"""

import logging

###########################
#    Public API Imports   #
###########################

from .version import version_info, __version__

from .policy import Policy, PolicyDeny, PolicyAllow

from .guard import (
    Inquiry,
    Guard,
)

from .effects import (
    ALLOW_ACCESS,
    DENY_ACCESS,
)

from .checker import (
    RegexChecker,
    StringFuzzyChecker,
    StringExactChecker,
    RulesChecker,
)

from . import rules

from .storage.memory import MemoryStorage

from .cache import (
    EnfoldCache,
    create_cached_guard
)


################
#  Setting up  #
################

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""
Utility functions and classes for Vakt.
"""

import logging
from abc import ABCMeta, abstractmethod

import jsonpickle


log = logging.getLogger(__name__)


class JsonSerializer:
    """
    Mixin for dumping object to JSON
    """

    @classmethod
    def from_json(cls, data):
        """
        Create object from a JSON string
        Returns a new instance of a class
        """
        return cls._parse(data)

    def to_json(self, sort=False):
        """
        Get JSON representation of an object
        """
        jsonpickle.set_encoder_options('json', sort_keys=sort)
        return jsonpickle.encode(self._data())

    @classmethod
    def _parse(cls, data):
        """Parse JSON string and return data"""
        try:
            return jsonpickle.decode(data)
        except ValueError as err:
            log.exception('Error creating %s from json.', cls.__name__)
            raise err

    def _data(self):
        """
        Get the object data. Is useful for overriding in custom classes
        """
        return vars(self)


class PrettyPrint:
    """
    Allows to log objects with all the fields
    """
    def __str__(self):
        return "%s <Object ID %s>: %s" % (self.__class__, id(self), vars(self))


class Subject:
    """
    Publisher of events in the pub-sub objects relation
    """
    def __init__(self):
        self._listeners = []

    def add_listener(self, listener):
        """
        Attach listener to this subject
        """
        self._listeners.append(listener)

    def remove_listener(self, listener):
        """
        Detach listener from this subject
        """
        self._listeners.remove(listener)

    def notify(self):
        """
        Notify all attached listeners about event
        """
        for listener in self._listeners:
            listener.update()


class Observer(metaclass=ABCMeta):
    """
    Observer of the events in the pub-sub objects relation
    """
    @abstractmethod
    def update(self):
        """
        Update observer on notify event
        """
        pass



import json
import uuid

from vakt.policy import Policy
from vakt.rules.operator import Eq
from vakt.rules.string import Equal
from vakt.storage.sql import PolicyModel, PolicySubjectModel, PolicyActionModel, PolicyResourceModel


def policy__model_assert(policy_model, policy):
    """
        Assert if the given policy model object `policy_model` is equal to the policy object `policy`
    """

    def element(el):
        if isinstance(el, dict):
            return json.dumps(el)
        elif isinstance(el, str):
            return el
        return None

    policy_json = policy.to_json()
    policy_dict = json.loads(policy_json)

    assert policy_model.uid == policy.uid
    assert policy_model.type == policy.type
    assert policy_model.description == policy.description

    assert len(policy_model.subjects) == len(policy.subjects)
    assert all(isinstance(policy_model.subjects[x], PolicySubjectModel) for x in range(len(policy_model.subjects)))
    assert all(policy_model.subjects[x].subject == element(policy_dict["subjects"][x]) for x in
               range(len(policy_model.subjects))) or \
           all(policy_model.subjects[x].subject_string == element(policy_dict["subjects"][x]) for x in
               range(len(policy_model.subjects)))

    assert len(policy_model.resources) == len(policy.resources)
    assert all(isinstance(policy_model.resources[x], PolicyResourceModel) for x in range(len(policy_model.resources)))
    assert all(policy_model.resources[x].resource == element(policy_dict["resources"][x]) for x in
               range(len(policy_model.resources))) or \
           all(policy_model.resources[x].resource_string == element(policy_dict["resources"][x]) for x in
               range(len(policy_model.resources)))

    assert len(policy_model.actions) == len(policy.actions)
    assert all(isinstance(policy_model.actions[x], PolicyActionModel) for x in range(len(policy_model.actions)))
    assert all(policy_model.actions[x].action == element(policy_dict["actions"][x]) for x in
               range(len(policy_model.actions))) or \
           all(policy_model.actions[x].action_string == element(policy_dict["actions"][x]) for x in
               range(len(policy_model.actions)))

    assert policy_model.context == json.dumps(policy_dict["context"])


class TestModel:

    def test_from_policy(self):
        id = str(uuid.uuid4())
        p = Policy(
            uid=id,
            description='foo bar баз',
            subjects=('Edward Rooney', 'Florence Sparrow'),
            actions=['<.*>'],
            resources=['<.*>'],
            context={
                'secret': Equal('i-am-a-teacher'),
            },
        )
        p_model = PolicyModel.from_policy(p)
        policy__model_assert(p_model, p)

    def test_update(self):
        id = str(uuid.uuid4())
        p = Policy(
            uid=id,
            description='foo bar баз',
            subjects=('Edward Rooney', 'Florence Sparrow'),
            actions=['<.*>'],
            resources=['<.*>'],
            context={
                'secret': Equal('i-am-a-teacher'),
            },
        )
        p_model = PolicyModel.from_policy(p)
        new_p = Policy(2, actions=[], subjects=[Eq('max'), Eq('bob')])
        p_model.update(new_p)
        policy__model_assert(p_model, new_p)

    def test_to_policy(self):
        id = str(uuid.uuid4())
        p = Policy(
            uid=id,
            description='foo bar баз',
            subjects=('Edward Rooney', 'Florence Sparrow'),
            actions=['<.*>'],
            resources=['<.*>'],
            context={
                'secret': Equal('i-am-a-teacher'),
            },
        )
        p_model = PolicyModel.from_policy(p)
        _p = p_model.to_policy()
        assert p.uid == _p.uid
        assert p.type == _p.type
        assert p.description == _p.description
        assert len(p.subjects) == len(_p.subjects)
        assert all(p.subjects[x] == _p.subjects[x] for x in range(len(p.subjects)))
        assert len(p.resources) == len(_p.resources)
        assert all(p.resources[x] == _p.resources[x] for x in range(len(p.resources)))
        assert len(p.actions) == len(_p.actions)
        assert all(p.actions[x] == _p.actions[x] for x in range(len(p.actions)))
        assert 'secret' in _p.context
        assert isinstance(_p.context['secret'], Equal)
        assert _p.context['secret'].val == 'i-am-a-teacher'



"""
MongoDB Storage and Migrations for Policies.
"""

import logging
import copy
from abc import ABCMeta

import bson.json_util as b_json
import pymongo
from pymongo.errors import DuplicateKeyError
import jsonpickle.tags

from ..storage.abc import Storage
from ..storage.migration import Migration, MigrationSet
from ..exceptions import PolicyExistsError, UnknownCheckerType, Irreversible
from ..policy import Policy
from ..rules.base import Rule
from ..checker import StringExactChecker, StringFuzzyChecker, RegexChecker, RulesChecker
from ..policy import TYPE_STRING_BASED, TYPE_RULE_BASED
from ..parser import compile_regex


DEFAULT_COLLECTION = 'vakt_policies'
DEFAULT_MIGRATION_COLLECTION = 'vakt_policies_migration_version'

log = logging.getLogger(__name__)


class MongoStorage(Storage):
    """Stores all policies in MongoDB"""

    def __init__(self, client, db_name, collection=DEFAULT_COLLECTION):
        self.client = client
        self.database = self.client[db_name]
        self.collection = self.database[collection]
        self.db_server_version = tuple(map(int, client.server_info()['version'].split('.')))
        self.condition_fields = [
            'actions',
            'subjects',
            'resources',
        ]
        self.condition_field_compiled_name = lambda x: '%s_compiled_regex' % x

    def add(self, policy):
        try:
            self.collection.insert_one(self.__prepare_doc(policy))
        except DuplicateKeyError:
            log.error('Error trying to create already existing policy with UID=%s.', policy.uid)
            raise PolicyExistsError(policy.uid)
        log.info('Added Policy: %s', policy)

    def get(self, uid):
        ret = self.collection.find_one(uid)
        if not ret:
            return None
        return self.__prepare_from_doc(ret)

    def get_all(self, limit, offset):
        self._check_limit_and_offset(limit, offset)
        # Special check for: https://docs.mongodb.com/manual/reference/method/cursor.limit/#zero-value
        if limit == 0:
            return []
        cur = self.collection.find(limit=limit, skip=offset, sort=[('_id', pymongo.ASCENDING)])
        return self.__feed_policies(cur)

    def find_for_inquiry(self, inquiry, checker=None):
        q_filter, use_aggregation = self._create_filter(inquiry, checker)
        if use_aggregation:
            cur = self.collection.aggregate(q_filter)
        else:
            cur = self.collection.find(q_filter)
        return self.__feed_policies(cur)

    def update(self, policy):
        uid = policy.uid
        self.collection.update_one(
            {'_id': uid},
            {"$set": self.__prepare_doc(policy)},
            upsert=False)
        log.info('Updated Policy with UID=%s. New value is: %s', uid, policy)

    def delete(self, uid):
        self.collection.delete_one({'_id': uid})
        log.info('Deleted Policy with UID=%s.', uid)

    def _create_filter(self, inquiry, checker):
        """
        Returns proper query-filter based on the checker type and a flag that marks whether aggregation should be used
        """
        if isinstance(checker, StringFuzzyChecker):
            return self.__string_query_on_conditions('$regex', inquiry), False
        elif isinstance(checker, StringExactChecker):
            return self.__string_query_on_conditions('$eq', inquiry), False
        elif isinstance(checker, RegexChecker):
            if self.db_server_version < (4, 2, 0):
                return {'type': TYPE_STRING_BASED}, False
            return self.__regex_query_on_conditions(inquiry), True
        elif isinstance(checker, RulesChecker):
            return {'type': TYPE_RULE_BASED}, False
        elif not checker:
            return {}, False
        else:
            log.error('Provided Checker type is not supported.')
            raise UnknownCheckerType(checker)

    def __string_query_on_conditions(self, operator, inquiry):
        """
        Construct MongoDB query for string-based Checkers.
        """
        conditions = [
            {'type': TYPE_STRING_BASED}
        ]
        for field in self.condition_fields:
            conditions.append({
                field: {
                    '$elemMatch': {
                        operator: getattr(inquiry, field.rstrip('s'))
                    }
                }
            })
        return {"$and": conditions}

    def __regex_query_on_conditions(self, inquiry):
        """
        Construct MongoDB query for RegexChecker.
        """
        conditions = [
            {
                '$eq': ['$type', TYPE_STRING_BASED]
            }
        ]
        for field in self.condition_fields:
            field_singular = field.rstrip('s')
            inquiry_value = getattr(inquiry, field_singular)
            conditions.append({
                '$anyElementTrue': [
                    {
                        '$map': {
                            'input': "$%s" % self.condition_field_compiled_name(field),
                            'as': field_singular,
                            'in': {
                                '$or': [
                                    {
                                        '$eq': ["$$%s" % field_singular, inquiry_value]
                                    },
                                    {
                                        '$regexMatch': {
                                            'input':  inquiry_value,
                                            'regex': "$$%s" % field_singular
                                        }
                                    }
                                ]
                            }
                        }
                    }
                ]
            })
        return [{'$match': {'$expr': {'$and': conditions}}}]

    def __prepare_doc(self, policy):
        """
        Prepare Policy object as a document for insertion.
        """
        # todo - add dict inheritance
        doc = b_json.loads(policy.to_json())
        if policy.type == TYPE_STRING_BASED:
            for field in self.condition_fields:
                compiled_regexes = []
                for el in doc[field]:
                    if policy.start_tag in el and policy.end_tag in el:
                        compiled = compile_regex(el, policy.start_tag, policy.end_tag).pattern
                    else:
                        compiled = el
                    compiled_regexes.append(compiled)
                doc[self.condition_field_compiled_name(field)] = compiled_regexes
        doc['_id'] = policy.uid
        return doc

    def __prepare_from_doc(self, doc):
        """
        Prepare Policy object as a return from MongoDB.
        """
        # todo - add dict inheritance
        del doc['_id']
        for field in self.condition_fields:
            compiled_field_name = self.condition_field_compiled_name(field)
            if compiled_field_name in doc:
                del doc[compiled_field_name]
        return Policy.from_json(b_json.dumps(doc))

    def __feed_policies(self, cursor):
        """
        Yields Policies from the given cursor.
        """
        for doc in cursor:
            yield self.__prepare_from_doc(doc)


##############
# Migrations #
##############

class MongoMigrationSet(MigrationSet):
    """
    Migrations Collection for MongoStorage
    """
    def __init__(self, storage, collection=DEFAULT_MIGRATION_COLLECTION):
        self.storage = storage
        self.collection = self.storage.database[collection]
        self.key = 'version'
        self.filter = {'_id': 'migration_version'}

    def migrations(self):
        return [
            Migration0To1x1x0(self.storage),
            Migration1x1x0To1x1x1(self.storage),
            Migration1x1x1To1x2x0(self.storage),
            Migration1x2x0To1x4x0(self.storage),
        ]

    def save_applied_number(self, number):
        self.collection.update_one(self.filter, {'$set': {self.key: number}}, upsert=True)

    def last_applied(self):
        data = self.collection.find_one(self.filter)
        if data:
            return int(data[self.key])
        return 0


class MongoMigration(Migration, metaclass=ABCMeta):
    """
    Mongo DB migration abstract base class
    """
    def _each_doc(self, processor):
        """
        Iterate each doc in the DB and run processor function with it
        """
        failed_policies = []
        storage = getattr(self, 'storage')
        cur = storage.collection.find()
        for doc in cur:
            try:
                log.info('Trying to migrate Policy with UID: %s', doc['uid'])
                new_doc = processor(doc)
                storage.collection.replace_one({'_id': new_doc['uid']}, new_doc)
                log.info('Policy with UID: %s was migrated', doc['uid'])
            except Irreversible as e:
                log.warning('Irreversible Policy. %s. Mongo doc: %s', e, doc)
                failed_policies.append(doc)
            except Exception as e:
                log.exception('Unexpected exception occurred while migrating Policy: %s', doc)
                failed_policies.append(doc)
        if failed_policies:
            msg = "\n".join([
                'Migration was unable to convert some Policies, but they were left in the database as-is. ' +
                'They might be not automatically convertible, custom ones, malformed JSON docs.',
                'You must convert them manually or delete entirely. See above log output for details of migration.',
                'Mongo IDs of failed Policies are: %s' % [p['_id'] for p in failed_policies]
            ])
            log.error(msg)


class Migration0To1x1x0(MongoMigration):
    """
    Migration between versions 0 and 1.1.0
    """

    def __init__(self, storage):
        self.storage = storage
        self.index_name = lambda i: i + '_idx'
        self.multi_key_indices = [
            'actions',
            'subjects',
            'resources',
        ]

    @property
    def order(self):
        return 1

    def up(self):
        # MongoDB automatically creates a multikey index if any indexed field is an array;
        # https://docs.mongodb.com/manual/core/index-multikey/#create-multikey-index
        for field in self.multi_key_indices:
            self.storage.collection.create_index(field, name=self.index_name(field))

    def down(self):
        for field in self.multi_key_indices:
            self.storage.collection.drop_index(self.index_name(field))


class Migration1x1x0To1x1x1(MongoMigration):
    """
    Migration between versions 1.1.0 and 1.1.1
    """

    def __init__(self, storage):
        self.storage = storage
        self._type_marker = jsonpickle.tags.OBJECT

    @property
    def order(self):
        return 2

    def up(self):
        def process(doc):
            """Processor for up"""
            doc_to_save = copy.deepcopy(doc)
            rules_to_save = {}
            for name, rule_str in doc['rules'].items():
                rule = b_json.loads(rule_str)
                rule_to_save = {self._type_marker: rule['type']}
                rule_to_save.update(rule['contents'])
                rules_to_save[name] = rule_to_save
            doc_to_save['rules'] = rules_to_save
            return doc_to_save
        self._each_doc(processor=process)

    def down(self):
        def process(doc):
            """Processor for down"""
            doc_to_save = copy.deepcopy(doc)
            rules_to_save = {}
            for name, rule in doc['rules'].items():
                rule_type = rule[self._type_marker]
                rule_contents = rule.copy()
                del rule_contents[self._type_marker]
                rule_to_save = {'type': rule_type, 'contents': {}}
                # check if we are dealing with 3-rd party or custom rules
                if not rule_type.startswith('vakt.rules.'):
                    for value in rule_contents.values():
                        # if rule has non-primitive data as its contents - we can't revert it to 1.1.0
                        if isinstance(value, (dict, Rule)) and jsonpickle.tags.RESERVED.intersection(value.keys()):
                            raise Irreversible('Custom rule class contains non-primitive data %s' % value)
                # vakt's own RegexMatchRule couldn't be stored in mongo because is has non-primitive data,
                # so it's impossible to put it to storage if we revert time back to 1.1.0
                elif rule_type == 'vakt.rules.string.RegexMatchRule':
                    raise Irreversible('vakt.rules.string.RegexMatchRule could not be stored in v1.1.0')
                rule_to_save['contents'].update(rule_contents)
                rules_to_save[name] = b_json.dumps(rule_to_save, sort_keys=True)
            # report or save document
            doc_to_save['rules'] = rules_to_save
            return doc_to_save
        self._each_doc(processor=process)


class Migration1x1x1To1x2x0(MongoMigration):
    """
    Migration between versions 1.1.1 and 1.2.0.
    What it does:
    - Adds index for `type` field
    - Updates Policies to:
        - have an appropriate type (all existing policies will become of a string-based type)
        - have 'context' attribute instead of 'rules' attribute
    """

    def __init__(self, storage):
        self.storage = storage
        self.type_field = 'type'
        self.type_index = 'type_idx'
        self.rules_rename = {
            'vakt.rules.string.StringEqualRule': 'vakt.rules.string.Equal',
            'vakt.rules.string.RegexMatchRule': 'vakt.rules.string.RegexMatch',
            'vakt.rules.string.StringPairsEqualRule': 'vakt.rules.string.PairsEqual',
            'vakt.rules.net.CIDRRule': 'vakt.rules.net.CIDR',
            'vakt.rules.inquiry.SubjectEqualRule': 'vakt.rules.inquiry.SubjectEqual',
            'vakt.rules.inquiry.ActionEqualRule': 'vakt.rules.inquiry.ActionEqual',
            'vakt.rules.inquiry.ResourceInRule': 'vakt.rules.inquiry.ResourceIn',
        }

    @property
    def order(self):
        return 3

    def up(self):
        def process(doc):
            """Processor for up"""
            doc['type'] = TYPE_STRING_BASED
            for rule in doc['rules'].values():
                rule_type = rule[jsonpickle.tags.OBJECT]
                for old, new in self.rules_rename.items():
                    if rule_type == old:
                        rule[jsonpickle.tags.OBJECT] = new
                        break
            doc['context'] = doc['rules']
            del doc['rules']
            return doc
        self.storage.collection.create_index(self.type_field, name=self.type_index)
        self._each_doc(processor=process)

    def down(self):
        def process(doc):
            """Processor for down"""
            if doc['type'] != TYPE_STRING_BASED:
                raise Irreversible('Policy is not of a string-based type, so not supported in < v1.2.0')
            for rule in doc['context'].values():
                rule_type = rule[jsonpickle.tags.OBJECT]
                for old, new in self.rules_rename.items():
                    if rule_type == new:
                        rule[jsonpickle.tags.OBJECT] = old
                        break
                if rule_type.startswith('vakt.rules.list') or \
                        rule_type.startswith('vakt.rules.logic') or \
                        rule_type.startswith('vakt.rules.operator') or \
                        rule_type in ['vakt.rules.string.StartsWith',
                                      'vakt.rules.string.EndsWith',
                                      'vakt.rules.string.Contains']:
                    raise Irreversible('Context contains rule that exist only in >= v1.2.0: %s' % rule)
            doc['rules'] = doc['context']
            del doc['context']
            del doc['type']
            return doc
        self.storage.collection.drop_index(self.type_index)
        self._each_doc(processor=process)


class Migration1x2x0To1x4x0(MongoMigration):
    """
    Migration between versions 1.2.0 and 1.4.0.
    What it does:
    - Adds special fields for each string-based policy that are needed for regex DB-side checks.
    - Adds indices for *_compiled_regex fields in string-based polices
    """

    def __init__(self, storage):
        self.storage = storage
        self.index_name = lambda i: i + '_idx'
        self.multi_key_indices = map(
            self.storage.condition_field_compiled_name,
            [
                'actions',
                'subjects',
                'resources',
            ]
        )

    @property
    def order(self):
        return 4

    def up(self):
        # create indices
        for field in self.multi_key_indices:
            self.storage.collection.create_index(field, name=self.index_name(field))
        # re-save policies to add compiled_regex fields
        for p in self.storage.retrieve_all():
            self.storage.update(p)

    def down(self):
        def process(doc):
            """Processor for down"""
            for field in [self.storage.condition_field_compiled_name(x) for x in self.storage.condition_fields]:
                if field in doc:
                    del doc[field]
            return doc
        # delete indices
        for field_name in self.multi_key_indices:
            self.storage.collection.drop_index(self.index_name(field_name))
        # return policies to their previous state
        self._each_doc(processor=process)



"""
SQL Storage for Policies.
"""

import logging

from sqlalchemy import and_, or_, literal, func
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import FlushError

from .model import PolicyModel, PolicyActionModel, PolicyResourceModel, PolicySubjectModel
from ..abc import Storage
from ...checker import StringExactChecker, StringFuzzyChecker, RegexChecker, RulesChecker
from ...exceptions import PolicyExistsError, UnknownCheckerType
from ...policy import TYPE_STRING_BASED, TYPE_RULE_BASED


log = logging.getLogger(__name__)


class SQLStorage(Storage):
    """Stores all policies in SQL Database"""

    def __init__(self, scoped_session):
        """
            Initialize SQL Storage

            :param scoped_session: SQL Alchemy scoped session
        """
        self.session = scoped_session
        self.dialect = self.session.bind.engine.dialect.name

    def add(self, policy):
        try:
            policy_model = PolicyModel.from_policy(policy)
            self.session.add(policy_model)
            self.session.commit()
        except IntegrityError:
            self.session.rollback()
            log.error('Error trying to create already existing policy with UID=%s.', policy.uid)
            raise PolicyExistsError(policy.uid)
        # todo - figure out why FlushError is raised instead of IntegrityError on PyPy tests
        except FlushError as e:
            if 'conflicts with persistent instance' in str(e):
                self.session.rollback()
                log.error('Error trying to create already existing policy with UID=%s.', policy.uid)
                raise PolicyExistsError(policy.uid)
        log.info('Added Policy: %s', policy)

    def get(self, uid):
        policy_model = self.session.query(PolicyModel).get(uid)
        if not policy_model:
            return None
        return policy_model.to_policy()

    def get_all(self, limit, offset):
        self._check_limit_and_offset(limit, offset)
        cur = self.session.query(PolicyModel).order_by(PolicyModel.uid.asc()).slice(offset, offset + limit)
        for policy_model in cur:
            yield policy_model.to_policy()

    def find_for_inquiry(self, inquiry, checker=None):
        cur = self._get_filtered_cursor(inquiry, checker)
        for policy_model in cur:
            yield policy_model.to_policy()

    def update(self, policy):
        try:
            policy_model = self.session.query(PolicyModel).get(policy.uid)
            if not policy_model:
                return
            policy_model.update(policy)
            self.session.commit()
        except IntegrityError:
            self.session.rollback()
            raise
        log.info('Updated Policy with UID=%s. New value is: %s', policy.uid, policy)

    def delete(self, uid):
        self.session.query(PolicyModel).filter(PolicyModel.uid == uid).delete()
        log.info('Deleted Policy with UID=%s.', uid)

    def _get_filtered_cursor(self, inquiry, checker):
        """
            Returns cursor with proper query-filter based on the checker type.
        """
        cur = self.session.query(PolicyModel)
        if isinstance(checker, StringFuzzyChecker):
            return cur.filter(
                PolicyModel.type == TYPE_STRING_BASED,
                PolicyModel.actions.any(PolicyActionModel.action_string.like('%{}%'.format(inquiry.action))),
                PolicyModel.resources.any(PolicyResourceModel.resource_string.like('%{}%'.format(inquiry.resource))),
                PolicyModel.subjects.any(PolicySubjectModel.subject_string.like('%{}%'.format(inquiry.subject))))
        elif isinstance(checker, StringExactChecker):
            return cur.filter(
                PolicyModel.type == TYPE_STRING_BASED,
                PolicyModel.actions.any(PolicyActionModel.action_string == inquiry.action),
                PolicyModel.resources.any(PolicyResourceModel.resource_string == inquiry.resource),
                PolicyModel.subjects.any(PolicySubjectModel.subject_string == inquiry.subject))
        elif isinstance(checker, RegexChecker):
            if not self._supports_regex_operator():
                return cur.filter(PolicyModel.type == TYPE_STRING_BASED)
            return cur.filter(
                PolicyModel.type == TYPE_STRING_BASED,
                PolicyModel.actions.any(
                    or_(
                        and_(PolicyActionModel.action_regex.is_(None),
                             PolicyActionModel.action_string == inquiry.action),
                        and_(PolicyActionModel.action_regex.isnot(None),
                             self._regex_operation(inquiry.action, PolicyActionModel.action_regex))
                    ),
                ),
                PolicyModel.resources.any(
                    or_(
                        and_(PolicyResourceModel.resource_regex.is_(None),
                             PolicyResourceModel.resource_string == inquiry.resource),
                        and_(PolicyResourceModel.resource_regex.isnot(None),
                             self._regex_operation(inquiry.resource, PolicyResourceModel.resource_regex))
                    ),
                ),
                PolicyModel.subjects.any(
                    or_(
                        and_(PolicySubjectModel.subject_regex.is_(None),
                             PolicySubjectModel.subject_string == inquiry.subject),
                        and_(PolicySubjectModel.subject_regex.isnot(None),
                             self._regex_operation(inquiry.subject, PolicySubjectModel.subject_regex))
                    ),
                )
            )
        elif isinstance(checker, RulesChecker):
            return cur.filter(PolicyModel.type == TYPE_RULE_BASED)
        elif not checker:
            return cur
        else:
            log.error('Provided Checker type is not supported.')
            raise UnknownCheckerType(checker)

    def _supports_regex_operator(self):
        """
        Does database support regex operator?
        """
        return self.dialect in ['mysql', 'postgresql', 'oracle']

    def _regex_operation(self, left, right):
        """
        Get database-specific regex operation.
        Don't forget to check if there is a support for regex operator before using it.
        """
        if self.dialect == 'mysql':
            return literal(left).op('REGEXP BINARY', is_comparison=True)(right)
        elif self.dialect == 'postgresql':
            return literal(left).op('~', is_comparison=True)(right)
        elif self.dialect == 'oracle':
            return func.REGEXP_LIKE(left, right)
        return None



from unittest.mock import Mock, patch

import pytest

from vakt.storage.memory import MemoryStorage
from vakt.storage.mongo import MongoStorage
from vakt import Policy, Inquiry, RulesChecker
from vakt.cache import EnfoldCache
from vakt.exceptions import PolicyExistsError
from ..helper import MemoryStorageYieldingExample2


class TestEnfoldCache:

    def test_init_without_populate(self):
        cache = MemoryStorage()
        policies = [Policy(1), Policy(2), Policy(3)]
        back = Mock(spec=MongoStorage, **{'retrieve_all.return_value': [policies]})
        c = EnfoldCache(back, cache=cache, populate=False)
        assert not back.retrieve_all.called
        assert [] == c.cache.get_all(1000, 0)

    def test_init_with_populate_and_populate_is_true_by_default(self):
        cache = MemoryStorage()
        policies = [Policy(1), Policy(2), Policy(3)]
        back = Mock(spec=MongoStorage, **{'retrieve_all.side_effect': [policies], 'get_all.side_effect': []})
        ec = EnfoldCache(back, cache=cache)
        assert back.retrieve_all.called
        assert policies == ec.cache.get_all(10000, 0)

    def test_init_with_populate_for_dirty_cache_storage(self):
        cache = MemoryStorage()
        cache.add(Policy(1))
        policies = [Policy(1), Policy(2), Policy(3)]
        back = Mock(spec=MongoStorage, **{'retrieve_all.side_effect': [policies]})
        with pytest.raises(Exception) as excinfo:
            EnfoldCache(back, cache=cache, populate=True)
        assert 'Conflicting UID = 1' == str(excinfo.value)
        assert back.retrieve_all.called

    def test_add_return_value(self):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        backend_return = back_storage.add(Policy(1))
        ec_return = ec.add(Policy(2))
        assert backend_return == ec_return

    def test_add_ok(self):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1)
        p2 = Policy(2)
        p3 = Policy(3)
        ec.add(p1)
        assert [p1] == cache_storage.get_all(100, 0)
        assert [p1] == back_storage.get_all(100, 0)
        ec.add(p2)
        ec.add(p3)
        assert [p1, p2, p3] == cache_storage.get_all(100, 0)
        assert [p1, p2, p3] == back_storage.get_all(100, 0)

    def test_add_fail(self):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1)
        p2 = Policy(2)
        p3 = Policy(3)
        ec.add(p1)
        back_storage.add = Mock(side_effect=PolicyExistsError('2'))
        with pytest.raises(PolicyExistsError) as excinfo:
            ec.add(p2)
        assert 'Conflicting UID = 2' == str(excinfo.value)
        assert [p1] == back_storage.get_all(1000, 0)
        # test general exception
        back_storage.add = Mock(side_effect=Exception('foo'))
        with pytest.raises(Exception) as excinfo:
            ec.add(p3)
        assert 'foo' == str(excinfo.value)
        assert [p1] == back_storage.get_all(1000, 0)
        assert [p1] == cache_storage.get_all(1000, 0)

    def test_update_return_value(self):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1, description='foo')
        ec.add(p1)
        p1.description = 'foo upd'
        backend_return = back_storage.update(p1)
        ec_return = ec.update(p1)
        assert backend_return == ec_return

    def test_update_ok(self):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1, description='foo')
        p2 = Policy(2, description='bar')
        p3 = Policy(3, description='baz')
        ec.add(p1)
        ec.add(p2)
        ec.add(p3)
        p1.description = 'foo2'
        ec.update(p1)
        p2.description = 'bar2'
        ec.update(p2)
        assert [p1, p2, p3] == cache_storage.get_all(100, 0)
        assert [p1, p2, p3] == back_storage.get_all(100, 0)

    def test_update_fail(self):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1, description='foo')
        p2 = Policy(2, description='bar')
        ec.add(p1)
        ec.add(p2)
        back_storage.update = Mock(side_effect=Exception('error!'))
        with pytest.raises(Exception) as excinfo:
            p1.description = 'changed foo'
            ec.update(p1)
        assert 'error!' == str(excinfo.value)
        assert [p1, p2] == back_storage.get_all(1000, 0)
        assert [p1, p2] == cache_storage.get_all(1000, 0)

    def test_delete_return_value(self):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1)
        ec.add(p1)
        backend_return = back_storage.delete(p1)
        ec_return = ec.delete(p1)
        assert backend_return == ec_return

    def test_delete_ok(self):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1, description='foo')
        p2 = Policy(2, description='bar')
        ec.add(p1)
        ec.add(p2)
        ec.delete(1)
        ec.delete(2)
        assert [] == cache_storage.get_all(100, 0)
        assert [] == back_storage.get_all(100, 0)

    def test_delete_fail(self):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1, description='foo')
        p2 = Policy(2, description='bar')
        ec.add(p1)
        ec.add(p2)
        back_storage.delete = Mock(side_effect=Exception('error!'))
        with pytest.raises(Exception) as excinfo:
            ec.delete(2)
        assert 'error!' == str(excinfo.value)
        assert [p1, p2] == back_storage.get_all(1000, 0)
        assert [p1, p2] == cache_storage.get_all(1000, 0)

    def test_get_return_value(self):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1)
        ec.add(p1)
        backend_return = back_storage.get(1)
        ec_return = ec.get(1)
        assert backend_return == ec_return

    @patch('vakt.cache.log')
    def test_get_for_non_populated_cache(self, log_mock):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        p1 = Policy(1, description='foo')
        p2 = Policy(2, description='bar')
        back_storage.add(p1)
        back_storage.add(p2)
        ec = EnfoldCache(back_storage, cache=cache_storage, populate=False)
        assert p1 == ec.get(1)
        log_mock.warning.assert_called_with(
            '%s cache miss for get Policy with UID=%s. Trying to get it from backend storage',
            'EnfoldCache',
            1
        )
        log_mock.reset_mock()
        assert p2 == ec.get(2)
        assert 1 == log_mock.warning.call_count
        log_mock.reset_mock()
        # test we won't return any inexistent policies
        assert ec.get(3) is None

    @patch('vakt.cache.log')
    def test_get_for_populated_cache(self, log_mock):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        p1 = Policy(1, description='foo')
        p2 = Policy(2, description='bar')
        back_storage.add(p1)
        back_storage.add(p2)
        ec = EnfoldCache(back_storage, cache=cache_storage, populate=True)
        assert p1 == ec.get(1)
        assert 0 == log_mock.warning.call_count
        log_mock.reset_mock()
        assert p2 == ec.get(2)
        assert 0 == log_mock.warning.call_count
        log_mock.reset_mock()
        # test we won't return any inexistent policies
        assert ec.get(3) is None
        log_mock.warning.assert_called_with(
            '%s cache miss for get Policy with UID=%s. Trying to get it from backend storage',
            'EnfoldCache',
            3
        )

    @pytest.mark.parametrize('storage', [
        MemoryStorage(),
        MemoryStorageYieldingExample2(),
    ])
    def test_get_all_return_value(self, storage):
        cache_storage = storage
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1)
        p2 = Policy(2)
        p3 = Policy(3)
        ec.add(p1)
        ec.add(p2)
        ec.add(p3)
        backend_return = back_storage.get_all(100, 0)
        ec_return = back_storage.get_all(100, 0)
        assert backend_return == ec_return

    @pytest.mark.parametrize('storage', [
        MemoryStorage(),
        MemoryStorageYieldingExample2(),
    ])
    def test_get_all(self, storage):
        p1 = Policy(1)
        p2 = Policy(2)
        p3 = Policy(3)
        p4 = Policy(4)
        p5 = Policy(5)
        cache_storage = storage
        back_storage = Mock(spec=MongoStorage, **{'get_all.return_value': []})
        ec = EnfoldCache(back_storage, cache=cache_storage, populate=False)
        assert [] == ec.get_all(1, 0)
        assert [] == ec.get_all(100, 0)
        # test we return from the backend
        back_storage.get_all = Mock(return_value=[p1, p2, p3])
        assert [p1, p2, p3] == list(ec.get_all(100, 0))
        back_storage.get_all = Mock(return_value=[])
        assert [] == list(ec.get_all(100, 0))
        # test we return from the cache
        cache_storage.add(p4)
        cache_storage.add(p5)
        back_storage.get_all = Mock(return_value=[p1, p2, p3])
        assert [p4] == list(ec.get_all(1, 0))
        assert [p5] == list(ec.get_all(1, 1))
        assert [p4, p5] == list(ec.get_all(2, 0))

    @pytest.mark.parametrize('storage', [
        MemoryStorage(),
        MemoryStorageYieldingExample2(),
    ])
    def test_retrieve_all_return_value(self, storage):
        cache_storage = storage
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        p1 = Policy(1)
        p2 = Policy(2)
        p3 = Policy(3)
        ec.add(p1)
        ec.add(p2)
        ec.add(p3)
        backend_return = back_storage.retrieve_all()
        ec_return = back_storage.retrieve_all()
        assert list(backend_return) == list(ec_return)

    @pytest.mark.parametrize('storage', [
        MemoryStorage(),
        MemoryStorageYieldingExample2(),
    ])
    def test_retrieve_all(self, storage):
        p1 = Policy(1)
        p2 = Policy(2)
        p3 = Policy(3)
        p4 = Policy(4)
        p5 = Policy(5)
        cache_storage = storage
        back_storage = Mock(spec=MongoStorage, **{'retrieve_all.return_value': []})
        ec = EnfoldCache(back_storage, cache=cache_storage, populate=False)
        assert [] == ec.retrieve_all()
        back_storage.retrieve_all.assert_called_with()
        assert [] == ec.retrieve_all(batch=10)
        back_storage.retrieve_all.assert_called_with(batch=10)
        assert [] == ec.retrieve_all(10)
        back_storage.retrieve_all.assert_called_with(10)
        # test we return from the backend
        back_storage.retrieve_all = Mock(return_value=[p1, p2, p3])
        assert [p1, p2, p3] == list(ec.retrieve_all())
        back_storage.retrieve_all = Mock(return_value=[])
        assert [] == list(ec.retrieve_all())
        # test we return from the cache (cache is dirty)
        cache_storage.add(p4)
        cache_storage.add(p5)
        back_storage.retrieve_all = Mock(return_value=[p1, p2, p3])
        assert [p4, p5] == list(ec.retrieve_all(1))
        assert [p4, p5] == list(ec.retrieve_all(batch=1))
        assert [p4, p5] == list(ec.retrieve_all())

    @pytest.mark.parametrize('storage', [
        MemoryStorage(),
        MemoryStorageYieldingExample2(),
    ])
    def test_find_for_inquiry_return_value(self, storage):
        cache_storage = storage
        back_storage = MemoryStorage()
        ec = EnfoldCache(back_storage, cache=cache_storage)
        inq = Inquiry(action='get')
        p1 = Policy(1)
        p2 = Policy(2)
        ec.add(p1)
        ec.add(p2)
        backend_return = back_storage.find_for_inquiry(inq)
        ec_return = ec.find_for_inquiry(inq)
        assert list(backend_return) == list(ec_return)

    @patch('vakt.cache.log')
    def test_find_for_inquiry_for_non_populated_cache(self, log_mock):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        inq = Inquiry(action='get')
        chk1 = RulesChecker()
        p1 = Policy(1, description='foo')
        p2 = Policy(2, description='bar')
        back_storage.add(p1)
        back_storage.add(p2)
        ec = EnfoldCache(back_storage, cache=cache_storage, populate=False)
        # Make first request
        assert [p1, p2] == list(ec.find_for_inquiry(inquiry=inq, checker=chk1))
        log_mock.warning.assert_called_with(
            '%s cache miss for find_for_inquiry. Trying it from backend storage', 'EnfoldCache'
        )
        log_mock.reset_mock()
        # Make second request
        assert [p1, p2] == list(ec.find_for_inquiry(inquiry=inq, checker=chk1))
        assert 1 == log_mock.warning.call_count
        log_mock.reset_mock()

    @patch('vakt.cache.log')
    def test_find_for_inquiry_for_populated_cache(self, log_mock):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        inq = Inquiry(action='get')
        chk1 = RulesChecker()
        p1 = Policy(1, description='foo')
        p2 = Policy(2, description='bar')
        cache_storage.add(p1)
        cache_storage.add(p2)
        ec = EnfoldCache(back_storage, cache=cache_storage, populate=True)
        # Make first request
        assert [p1, p2] == list(ec.find_for_inquiry(inquiry=inq, checker=chk1))
        assert 0 == log_mock.warning.call_count
        log_mock.reset_mock()
        # Make second request
        assert [p1, p2] == list(ec.find_for_inquiry(inquiry=inq, checker=chk1))
        assert 0 == log_mock.warning.call_count

    @patch('vakt.cache.log')
    def test_general_flow(self, log_mock):
        cache_storage = MemoryStorage()
        back_storage = MemoryStorage()
        inq = Inquiry(action='get')
        chk1 = RulesChecker()
        p1 = Policy(1, description='initial')
        p2 = Policy(2, description='initial')
        p3 = Policy(3, description='added later')
        # initialize backend storage
        back_storage.add(p1)
        back_storage.add(p2)
        # create enfold-cache but do not populate it
        ec = EnfoldCache(back_storage, cache=cache_storage, populate=False)
        # make sure policies are returned from the backend
        assert [p1, p2] == list(ec.find_for_inquiry(inquiry=inq, checker=chk1))
        # make sure we logged warning about cache miss
        log_mock.warning.assert_called_with(
            '%s cache miss for find_for_inquiry. Trying it from backend storage', 'EnfoldCache'
        )
        log_mock.reset_mock()
        # populate cache with backend policies so that we do not make cache hit misses
        ec.populate()
        # make sure policies are returned
        assert [p1, p2] == list(ec.find_for_inquiry(inquiry=inq, checker=chk1))
        # make sure we do not have cache misses
        assert 0 == log_mock.warning.call_count
        log_mock.reset_mock()
        # let's add a new policy via enfold-cache
        ec.add(p3)
        # make sure policies are returned
        assert [p1, p2, p3] == list(ec.find_for_inquiry(inquiry=inq, checker=chk1))
        # make sure we do not have cache misses
        assert 0 == log_mock.warning.call_count
        log_mock.reset_mock()
        # make sure we have those policies in the backend-storage
        assert [p1, p2, p3] == list(back_storage.find_for_inquiry(inquiry=inq, checker=chk1))
        # make sure we have those policies in the cache-storage
        assert [p1, p2, p3] == list(cache_storage.find_for_inquiry(inquiry=inq, checker=chk1))
        # -----------------------
        # -----------------------
        # -----------------------
        # let's re-create enfold cache. This time with initial population
        cache_storage2 = MemoryStorage()
        ec2 = EnfoldCache(back_storage, cache=cache_storage2, populate=True)
        # make sure we have all policies in the cache-storage
        assert [p1, p2, p3] == list(cache_storage2.find_for_inquiry(inquiry=inq, checker=chk1))
        # make sure policies are returned by find_for_inquiry
        assert [p1, p2, p3] == list(ec2.find_for_inquiry(inquiry=inq, checker=chk1))
        # make sure we do not have cache misses
        assert 0 == log_mock.warning.call_count
        log_mock.reset_mock()



"""
Module for various checkers.
"""

import re
import logging
from functools import lru_cache
from abc import ABCMeta, abstractmethod

from .parser import compile_regex
from .exceptions import InvalidPatternError


log = logging.getLogger(__name__)


class Checker(metaclass=ABCMeta):
    """
    Abstract class for Checker typing.
    """
    @abstractmethod
    def fits(self, policy, field, what, inquiry=None):
        """
        Check if fields from Inquiry fit some Policies
        """
        pass


class RegexChecker(Checker):
    """
    Checker that uses regular expressions.
    E.g. 'Dog', 'Doge', 'Dogs' fit <Dog[se]?>
         'Dogger' doesn't fit <Dog[se]?>
    """

    def __init__(self, cache_size=1024):
        """Set up LRU-cache size for compiled regular expressions."""
        self.compile = lru_cache(maxsize=cache_size)(compile_regex)

    def fits(self, policy, field, what, inquiry=None):
        """Does Policy fit the given 'what' value by its 'field' property"""
        where = getattr(policy, field, [])
        for i in where:
            # We are not meant to handle non-string values if they accidentally got here
            if type(i) != str:
                continue
            # check if 'where' item is not written in a policy-defined-regex syntax.
            if policy.start_tag not in i and policy.end_tag not in i:
                if i != what:
                    continue       # continue if it's not a string match
                else:
                    return True    # we've found a string match - policy fits by simple string value
            try:
                pattern = self.compile(i, policy.start_tag, policy.end_tag)
            except InvalidPatternError:
                log.exception('Error matching policy, because of failed regex %s compilation', i)
                return False
            if re.match(pattern, what):
                return True
        return False


class StringChecker(Checker):
    """
    Checker that uses string equality.
    You have to redefine `compare` method.
    """

    def fits(self, policy, field, what, inquiry=None):
        """Does Policy fit the given 'what' value by its 'field' property"""
        where = getattr(policy, field, [])
        for item in where:
            # We are not meant to handle non-string values if they accidentally got here
            if type(item) != str:
                continue
            if policy.start_tag == item[0] and policy.end_tag == item[-1]:
                item = item[1:-1]
            if self.compare(what, item):
                return True
        return False

    @abstractmethod
    def compare(self, needle, haystack):
        """Compares two string values. Override it in a subclass"""
        pass


class StringExactChecker(StringChecker):
    """
    Checker that uses exact string equality. Case-sensitive.
    E.g. 'sun' in 'sunny' - False
         'sun' in 'sun' - True
    """

    def compare(self, needle, haystack):
        return needle == haystack


class StringFuzzyChecker(StringChecker):
    """
    Checker that uses fuzzy substring equality. Case-sensitive.
    E.g. 'sun' in 'sunny' - True
         'sun' in 'unsung' - True
         'sun' in 'sun' - True
    """

    def compare(self, needle, haystack):
        return needle in haystack


class RulesChecker(Checker):
    """
    Checker that uses Rules defined inside dictionaries to determine match.
    """
    def fits(self, policy, field, what, inquiry=None):
        """Does Policy fit the given 'what' value by its 'field' property"""
        where_list = getattr(policy, field, [])
        is_what_dict = isinstance(what, dict)
        for i in where_list:
            item_result = False
            # If not dict or Rule, skip it - we are not meant to handle it.
            # Do not use isinstance for higher execution speed
            if type(i) == dict:
                for key, rule in i.items():
                    if not is_what_dict:
                        log.debug('Error matching Policy: data %r in Inquiry is not `dict`', what)
                        item_result = False
                    # at least one missing key in inquiry's data means no match for this item
                    elif key not in what:
                        log.debug('Error matching Policy: data %r has no key "%r" required by Policy', what, key)
                        item_result = False
                    else:
                        what_value = what[key]
                        item_result = self._check_satisfied(rule, what_value, inquiry)
                    # at least one item's key didn't satisfy -> fail fast: policy doesn't fit anyway
                    if not item_result:
                        break
            elif callable(getattr(i, 'satisfied', '')):
                item_result = self._check_satisfied(i, what, inquiry)
            # If at least one item fits -> policy fits for this field
            if item_result:
                return True
        return False

    @staticmethod
    def _check_satisfied(rule, what_value, inquiry=None):
        try:
            return rule.satisfied(what_value, inquiry)
        # broad exception for possible custom exceptions. Any exception -> no match
        # todo - decide on granular handler
        except Exception:
            log.exception('Error matching Policy, because of raised exception')
            return False



"""
Exceptions relevant for Vakt workflow.
"""


class InvalidPatternError(Exception):
    """Invalid policy pattern"""
    pass


class PolicyCreationError(Exception):
    """Error during Policy creation occurred."""
    pass


class RuleCreationError(Exception):
    """Error during Rule creation occurred."""
    pass


class PolicyExistsError(Exception):
    """Error when the already existing policy is attempted to be created by Storage"""
    def __init__(self, uid):
        super().__init__('Conflicting UID = %s' % uid)


class UnknownCheckerType(Exception):
    """Storage can't determine Checker type based on provided one."""
    def __init__(self, obj):
        super().__init__("Can't determine Checker type. Given: %s" % type(obj).__name__)


class Irreversible(Exception):
    """Storage migration can't convert record back to a lower version."""
    pass



import pytest

from vakt.policy import Policy, PolicyAllow, PolicyDeny
from vakt.effects import ALLOW_ACCESS, DENY_ACCESS
from vakt.exceptions import PolicyCreationError
from vakt.rules.net import CIDR
from vakt.rules.string import Equal
from vakt.rules.operator import Eq, Greater
from vakt.rules.logic import And, Any
from vakt.rules.list import AnyIn
from vakt.policy import TYPE_STRING_BASED, TYPE_RULE_BASED


def test_types():
    assert 1 == TYPE_STRING_BASED
    assert 2 == TYPE_RULE_BASED


def test_properties():
    policy = Policy('123', description='readme',
                    subjects=['user'], effect=ALLOW_ACCESS,
                    resources=r'books:{\d+}', actions=['create', 'delete'], context={})
    assert '123' == policy.uid
    assert 'readme' == policy.description
    assert ['user'] == policy.subjects
    assert ALLOW_ACCESS == policy.effect
    assert r'books:{\d+}' == policy.resources
    assert ['create', 'delete'] == policy.actions
    assert {} == policy.context


def test_exception_raised_when_context_is_not_dict():
    with pytest.raises(PolicyCreationError):
        Policy('1', context=[Equal('foo')])


@pytest.mark.parametrize('data, expect', [
    ('{"uid":123}',
     '{"actions": [], "context": {}, "description": null, "effect": "deny", ' +
     '"resources": [], "subjects": [], "type": 1, "uid": 123}'),
    ('{"effect":"allow", "actions": ["create", "update"], "uid":123}',
     '{"actions": ["create", "update"], "context": {}, "description": null, "effect": "allow", ' +
     '"resources": [], "subjects": [], "type": 1, "uid": 123}'),
    # 'type' if present, should be omitted and not result in setting type of a Policy object
    ('{"actions": ["create", "update"], "uid":123, "type": 2}',
     '{"actions": ["create", "update"], "context": {}, "description": null, "effect": "deny", ' +
     '"resources": [], "subjects": [], "type": 1, "uid": 123}'),
    # 'rules' can be present, but after deserialization should be converted to 'context' field
    ('{"uid":123, "actions": [], "rules":' +
     '{"a": {"py/object": "vakt.rules.net.CIDR", "cidr": "192.168.2.0/24"}}}',
     '{"actions": [], "context": {"a": {"cidr": "192.168.2.0/24", "py/object": "vakt.rules.net.CIDR"}}, ' +
     '"description": null, "effect": "deny", "resources": [], "subjects": [], "type": 1, "uid": 123}'),
    # 'context' should be deserialized to 'context' properly
    ('{"uid":123, "actions": [], "rules":' +
     '{"a": {"py/object": "vakt.rules.net.CIDR", "cidr": "192.168.2.0/24"}}}',
     '{"actions": [], "context": {"a": {"cidr": "192.168.2.0/24", "py/object": "vakt.rules.net.CIDR"}}, ' +
     '"description": null, "effect": "deny", "resources": [], "subjects": [], "type": 1, "uid": 123}'),
    # 'context' should win over deprecated attribute 'rules' if both are present
    ('{"uid":123, "actions": [], "context": {"a": {"py/object": "vakt.rules.string.Equal", "val": "foo"}}, ' +
     '"rules": {"b": {"cidr": "192.168.2.0/24", "py/object": "vakt.rules.net.CIDR"}}}',
     '{"actions": [], "context": {"a": {"py/object": "vakt.rules.string.Equal", "val": "foo"}}, ' +
     '"description": null, "effect": "deny", "resources": [], "subjects": [], "type": 1, "uid": 123}'),
])
def test_json_roundtrip(data, expect):
    p = Policy.from_json(data)
    assert expect == p.to_json(sort=True)


@pytest.mark.parametrize('data, effect', [
    ('{"uid":123}', DENY_ACCESS),
    ('{"uid":123, "effect":""}', DENY_ACCESS),
    ('{"uid":123, "effect":"deny"}', DENY_ACCESS),
    ('{"uid":123, "effect":null}', DENY_ACCESS),
    ('{"uid":123, "effect":"allow"}', ALLOW_ACCESS),
])
def test_json_default_effect_is_set_correctly_when_from_json(data, effect):
    p = Policy.from_json(data)
    assert effect == p.effect


def test_json_roundtrip_of_a_policy_with_context():
    p = Policy('123', context={'ip': CIDR('192.168.1.0/24'), 'sub': Equal('test-me')})
    s = p.to_json()
    p1 = Policy.from_json(s)
    assert '123' == p1.uid
    assert 2 == len(p1.context)
    assert 'ip' in p1.context
    assert 'sub' in p1.context
    assert isinstance(p1.context['ip'], CIDR)
    assert isinstance(p1.context['sub'], Equal)
    assert p1.context['sub'].satisfied('test-me')

    # 'context' wins over deprecated rules
    p = Policy(
        '456',
        context={'ip': CIDR('192.168.1.0/24'), 'sub': Equal('foo-bar')},
        rules={'ip': CIDR('127.0.0.1'), 'sub': Equal('baz')}
    )
    s = p.to_json()
    p1 = Policy.from_json(s)
    assert '456' == p1.uid
    assert 2 == len(p1.context)
    assert 'ip' in p1.context
    assert 'sub' in p1.context
    assert isinstance(p1.context['ip'], CIDR)
    assert isinstance(p1.context['sub'], Equal)
    assert p1.context['sub'].satisfied('foo-bar')
    assert p1.context['ip'].satisfied('192.168.1.0')
    assert not hasattr(p1, 'rules')

    # 'rules' are allowed, but they become a 'context' class field
    with pytest.deprecated_call():
        p = Policy('789', rules={'ip': CIDR('127.0.0.1'), 'sub': Equal('baz')})
        s = p.to_json()
        p1 = Policy.from_json(s)
        assert '789' == p1.uid
        assert 2 == len(p1.context)
        assert 'ip' in p1.context
        assert 'sub' in p1.context
        assert isinstance(p1.context['ip'], CIDR)
        assert isinstance(p1.context['sub'], Equal)
        assert p1.context['sub'].satisfied('baz')
        assert p1.context['ip'].satisfied('127.0.0.1')
        assert not hasattr(p1, 'rules')


@pytest.mark.parametrize('policy', [
    Policy(1, subjects=[{'name': Eq('Max'), 'rate': Greater(90)}], actions=[Eq('get'), Eq('post')], resources=[Any()]),
    Policy(2, subjects=[{'login': Eq('sally')}], actions=[Eq('get'), Eq('post')], context={'ip': Eq('127.0.0.1')}),
    Policy(3, subjects=[{'rating': AnyIn(1, 2)}], actions=[And(Eq('get'), Eq('post'))]),
    Policy(4, subjects=[{'rating': AnyIn(1, 2)}], actions=[And(Eq('get'), Eq('post'))]),
    Policy(5, actions=[Eq('get')]),
])
def test_json_roundtrip_of_a_rules_based_policy(policy):
    pj = policy.to_json()
    p2 = Policy.from_json(pj)
    assert policy.to_json() == p2.to_json()


@pytest.mark.parametrize('data, exception, msg', [
    ('{}', PolicyCreationError, "'uid'"),
    ('{"uid":}', ValueError, ''),
    ('', ValueError, ''),
])
def test_json_roundtrip_not_create_policy(data, exception, msg):
    with pytest.raises(exception) as excinfo:
        Policy.from_json(data)
    assert msg in str(excinfo.value)


@pytest.mark.parametrize('effect, result', [
    ('foo', False),
    ('', False),
    (None, False),
    (DENY_ACCESS, False),
    (ALLOW_ACCESS, True),
])
def test_allow_access(effect, result):
    p = Policy('1', effect=effect)
    assert result == p.allow_access()


def test_start_tag():
    p = Policy('1')
    assert '<' == p.start_tag


def test_end_tag():
    p = Policy('1')
    assert '>' == p.end_tag


def test_pretty_print():
    p = Policy('1', description='readme', subjects=['user'])
    assert "<class 'vakt.policy.Policy'>" in str(p)
    assert "'uid': '1'" in str(p)
    assert "'description': 'readme'" in str(p)
    assert "'subjects': ['user']" in str(p)
    assert "'effect': 'deny'" in str(p)
    assert "'resources': ()" in str(p)
    assert "'actions': ()" in str(p)
    assert "'context': {}" in str(p)
    p = Policy('2', actions=[Eq('get'), Eq('post')])
    assert "vakt.rules.operator.Eq" in str(p.actions)


@pytest.mark.parametrize('policy, policy_type', [
    (Policy(1), TYPE_STRING_BASED),
    (Policy(1, actions=['<foo.bar>']), TYPE_STRING_BASED),
    (Policy(1, actions=[], resources=[], subjects=[], context={}), TYPE_STRING_BASED),
    (Policy(1, actions=['<foo.bar>'], resources=['asdf']), TYPE_STRING_BASED),
    (Policy(1, actions=['<foo.bar>'], resources=['asdf'], subjects=['<qwerty>']), TYPE_STRING_BASED),
    (Policy(1, actions=['<foo.bar>'], resources=['asdf'], subjects=['<qwerty>'], context={}), TYPE_STRING_BASED),
    (Policy(1, actions=['books:<foo.bar>'], resources=['asdf'], subjects=['<qwerty>'], context={}), TYPE_STRING_BASED),
    (Policy(1, actions=[{'ip': CIDR('127.0.0.1')}]), TYPE_RULE_BASED),
    (Policy(1, actions=[Eq('get')]), TYPE_RULE_BASED),
    (Policy(1, actions=[Eq('get'), {'read': Eq(True)}]), TYPE_RULE_BASED),
    (Policy(1, actions=[{'ip': CIDR('127.0.0.1')}, {'ip': CIDR('10.12.35.88')}]), TYPE_RULE_BASED),
    (Policy(1, actions=[{'ip': CIDR('127.0.0.1')}], resources=[{'ip': CIDR('127.0.0.1')}]), TYPE_RULE_BASED),
    (Policy(1, actions=[{'ip': CIDR('127.0.0.1')}],
            resources=[{'ip': CIDR('127.0.0.1')}], subjects=[{'ip': CIDR('0.0.0.0')}]), TYPE_RULE_BASED),
    (Policy(1, subjects=[{'name': Eq('Max'), 'rate': Greater(90)}], actions=[Eq('get'), Eq('post')],
            resources=[{'ip': CIDR('127.0.0.1')}]), TYPE_RULE_BASED)
])
def test_policy_type_on_creation(policy, policy_type):
    assert policy_type == policy.type


@pytest.mark.parametrize('policy_data', [
    {'uid': 1, 'actions': [{'ip': CIDR('127.0.0.1')}, '<.*>']},
    {'uid': 1, 'resources': [{'ip': CIDR('127.0.0.1')}, '<.*>']},
    {'uid': 1, 'subjects': [{'ip': CIDR('127.0.0.1')}, '<.*>']},
    {'uid': 1, 'subjects': [{'ip': CIDR('127.0.0.1')}, '<.*>'], 'resources': ['asdf']},
    {'uid': 1, 'subjects': [{'ip': CIDR('127.0.0.1')}], 'resources': ['<asdf>']},
    {'uid': 1, 'subjects': [{'ip': CIDR('127.0.0.1')}], 'actions': ['foo']},
    {'uid': 1, 'subjects': [Eq('Molly')], 'actions': ['foo']},
    {'uid': 1, 'subjects': ['Jane'], 'actions': [Eq('run')]},
])
def test_policy_raises_exception_if_mixed_elements(policy_data):
    with pytest.raises(PolicyCreationError):
        Policy(**policy_data)


def test_policy_type_on_attribute_change():
    p = Policy(1, actions=['<foo.bar>'], resources=['asdf'], subjects=['<qwerty>'])
    assert TYPE_STRING_BASED == p.type
    p.effect = ALLOW_ACCESS
    assert TYPE_STRING_BASED == p.type
    with pytest.raises(PolicyCreationError):
        p.actions = [{'ip': CIDR('0.0.0.0')}]
    assert TYPE_STRING_BASED == p.type
    with pytest.raises(PolicyCreationError):
        p.subjects = [{'ip': CIDR('0.0.0.0')}]
    with pytest.raises(PolicyCreationError):
        p.actions = [Any()]
    assert TYPE_STRING_BASED == p.type
    p.actions = ['<.*>']
    assert TYPE_STRING_BASED == p.type
    p.subjects = ['<.*>']
    assert TYPE_STRING_BASED == p.type
    p.type = TYPE_RULE_BASED  # explicit assign doesn't help
    assert TYPE_STRING_BASED == p.type
    # testing the from the opposite direction
    p = Policy(2, actions=[Any()], resources=[{'book': Eq('UX Manual')}], subjects=[Eq('Sally'), Eq('Bob')])
    assert TYPE_RULE_BASED == p.type
    p.effect = ALLOW_ACCESS
    assert TYPE_RULE_BASED == p.type
    with pytest.raises(PolicyCreationError):
        p.actions = ['<foo.bar>']
    assert TYPE_RULE_BASED == p.type
    with pytest.raises(PolicyCreationError):
        p.subjects = ['<foo.bar>', 'baz']
    with pytest.raises(PolicyCreationError):
        p.actions = ['baz<.*>']
    assert TYPE_RULE_BASED == p.type
    p.actions = [Any()]
    assert TYPE_RULE_BASED == p.type
    p.subjects = [Any()]
    assert TYPE_RULE_BASED == p.type
    p.type = TYPE_STRING_BASED  # explicit assign doesn't help
    assert TYPE_RULE_BASED == p.type


@pytest.mark.parametrize('args, msg', [
    ({'actions': (1, 2)}, 'Field "actions" element must be of `str`, `dict` or `Rule` type.'),
    ({'actions': (1, 'abc')}, 'Field "actions" element must be of `str`, `dict` or `Rule` type.'),
    ({'actions': ('2', 1)}, 'Field "actions" element must be of `str`, `dict` or `Rule` type.'),
    ({'actions': ([3])}, 'Field "actions" element must be of `str`, `dict` or `Rule` type.'),
    ({'actions': ([], [])}, 'Field "actions" element must be of `str`, `dict` or `Rule` type.'),
    ({'subjects': (1, {})}, 'Field "subjects" element must be of `str`, `dict` or `Rule` type'),
    ({'context': ()}, 'Error creating Policy. Context must be a dictionary'),
    ({'context': 5}, 'Error creating Policy. Context must be a dictionary'),
    ({'context': 'data'}, 'Error creating Policy. Context must be a dictionary'),
])
def test_policy_field_type_check(args, msg):
    with pytest.raises(PolicyCreationError) as excinfo:
        Policy(1, **args)
    assert msg in str(excinfo.value)


@pytest.mark.parametrize('klass, is_allowed, effect', [
    (PolicyAllow, True, 'allow'),
    (PolicyDeny, False, 'deny'),
])
def test_PolicyAllow_and_PolicyDeny(klass, is_allowed, effect):
    p = klass(1, actions=['<foo.bar>'], resources=['asdf'],
              subjects=['<qwerty>'], description='test')
    assert is_allowed == p.allow_access()
    assert 1 == p.uid
    assert 'test' == p.description
    assert TYPE_STRING_BASED == p.type
    assert ['<foo.bar>'] == p.actions
    assert ['asdf'] == p.resources
    assert ['<qwerty>'] == p.subjects
    assert {} == p.context
    assert '{"actions": ["<foo.bar>"], "context": {}, "description": "test", "effect": "%s", ' % effect + \
           '"resources": ["asdf"], "subjects": ["<qwerty>"], "type": 1, "uid": 1}' == p.to_json(sort=True)
    assert ['<foo.bar>'] == Policy.from_json(p.to_json()).actions
    p.effect = DENY_ACCESS
    assert DENY_ACCESS == p.effect
    p2 = klass(2, context={'a': Eq(100)})
    assert isinstance(p2.context.get('a'), Eq)
    assert 100 == p2.context.get('a').val
    # check positional arguments
    p3 = Policy(1, actions=['<foo.bar>'], resources=['asdf'],
                subjects=['<qwerty>'], description='test', effect=ALLOW_ACCESS if is_allowed else DENY_ACCESS)
    p4 = klass(1, ['<qwerty>'], ['asdf'], ['<foo.bar>'], {}, 'test')
    assert p3.to_json(sort=True) == p4.to_json(sort=True)



"""
Memory storage for Policies.
"""

import threading
import logging

from ..storage.abc import Storage
from ..exceptions import PolicyExistsError


log = logging.getLogger(__name__)


class MemoryStorage(Storage):
    """Stores all policies in memory"""

    def __init__(self):
        self.policies = {}
        self.lock = threading.Lock()

    def add(self, policy):
        uid = policy.uid
        with self.lock:
            if uid in self.policies:
                log.error('Error trying to create already existing policy with UID=%s', uid)
                raise PolicyExistsError(uid)
            self.policies[uid] = policy
            log.info('Added Policy: %s', policy)

    def get(self, uid):
        return self.policies.get(uid)

    def get_all(self, limit, offset):
        self._check_limit_and_offset(limit, offset)
        result = [v for v in self.policies.values()]
        if offset > len(result) or limit == 0:
            return []
        return result[offset:limit+offset]

    def find_for_inquiry(self, inquiry, checker=None):
        with self.lock:
            return self.policies.values()

    def update(self, policy):
        self.policies[policy.uid] = policy
        log.info('Updated Policy with UID=%s. New value is: %s', policy.uid, policy)

    def delete(self, uid):
        if uid in self.policies:
            del self.policies[uid]
            log.info('Policy with UID %s was deleted', uid)
